# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import List, Union
from .utils import read_description_from_odx
from .structures import BasicStructure
from .dataobjectproperty import DopBase
from .decodestate import DecodeState
from .encodestate import EncodeState


class EndOfPduField(DopBase):
    """ End of PDU fields are structures that are repeated until the end of the PDU """

    def __init__(self,
                 id,
                 short_name,
                 structure=None,
                 structure_ref=None,
                 structure_snref=None,
                 min_number_of_items=0,
                 max_number_of_items=None,
                 is_visible=False,
                 long_name=None,
                 description=None):
        super().__init__(id, short_name, long_name=long_name,
                         description=description, is_visible=is_visible)

        self.structure_snref = structure_snref
        self.structure_ref = structure_ref

        self._structure = structure
        if structure:
            self.structure_ref = structure.id
        assert self.structure_ref or self.structure_snref

        self.min_number_of_items = min_number_of_items
        self.max_number_of_items = max_number_of_items

        self.is_visible = is_visible

    @property
    def structure(self) -> BasicStructure:
        """may be a Structure or a env-data-desc"""
        return self._structure

    @property
    def bit_length(self):
        return self.structure.bit_length

    def convert_physical_to_internal(self, physical_value):
        return self.structure.convert_physical_to_internal(physical_value)

    def convert_physical_to_bytes(self, physical_value: Union[dict, List[dict]], encode_state: EncodeState, bit_position=0):
        assert bit_position == 0, "End of PDU field must be byte aligned. Is there an error in reading the .odx?"
        if isinstance(physical_value, dict):
            # If the value is given as a dict, the End of PDU field behaves like the underlying structure.
            return self.structure.convert_physical_to_bytes(physical_value, encode_state)
        else:
            assert isinstance(physical_value, List), "The value of an End of PDU field must be a list or a dict." 
            # If the value is given as a list, each list element is a encoded seperately using the structure.
            coded_rpc = bytes()
            for value in physical_value:
                encode_state = encode_state._replace(coded_message=bytes())
                coded_rpc += self.structure.convert_physical_to_bytes(value, encode_state)
            return coded_rpc

    def convert_bytes_to_physical(self, decode_state: DecodeState, bit_position=0):
        next_byte_position = decode_state.next_byte_position
        byte_code = decode_state.coded_message

        value = []
        while len(byte_code) > next_byte_position:
            # ATTENTION: the ODX specification is very misleading
            # here: it says that the item is repeated until the end of
            # the PDU, but it means that DOP of the items that are
            # repeated are identical, not their values
            new_value, next_byte_position = self.structure.convert_bytes_to_physical(decode_state,
                                                                                     bit_position=bit_position)
            # Update next byte_position
            decode_state = decode_state._replace(next_byte_position=next_byte_position)
            value.append(new_value)

        return value, next_byte_position

    def _resolve_references(self, parent_dl, id_lookup):
        """Recursively resolve any references (odxlinks or sn-refs)
        """
        if self.structure_ref is not None:
            self._structure = id_lookup[self.structure_ref]
        else:
            self._structure = parent_dl.data_object_properties[self.structure_snref]

    def __repr__(self) -> str:
        return f"EndOfPduField(short_name='{self.short_name}', ref='{self.structure.id}')"

    def __str__(self):
        return "\n".join([
            f"EndOfPduField(short_name='{self.short_name}', ref='{self.structure.id}')"
        ] + [
            " " + str(self.structure).replace("\n", "\n ")
        ])


def read_end_of_pdu_field_from_odx(et_element):
    id = et_element.get("ID")
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.find(
        "LONG-NAME").text if et_element.find("LONG-NAME") is not None else None
    description = read_description_from_odx(et_element.find("DESC"))

    if et_element.find("BASIC-STRUCTURE-REF") is not None:
        structure_ref = et_element.find(
            "BASIC-STRUCTURE-REF").get("ID-REF")
        structure_snref = None

    if et_element.find("BASIC-STRUCTURE-SNREF") is not None:
        structure_ref = None
        structure_snref = et_element.find(
            "BASIC-STRUCTURE-SNREF").get("SHORT-NAME")

    if et_element.find("ENV-DATA-DESC-REF") is not None:
        structure_ref = et_element.get("ENV-DATA-DESC-REF")
        structure_snref = None

    if et_element.find("ENV-DATA-DESC-SNREF") is not None:
        structure_ref = None
        structure_snref = et_element.get("ENV-DATA-DESC-SNREF")

    if et_element.find("MIN-NUMBER-OF-ITEMS") is not None:
        min_number_of_items = int(
            et_element.find("MIN-NUMBER-OF-ITEMS").text)
    else:
        min_number_of_items = None
    if et_element.find("MAX-NUMBER-OF-ITEMS") is not None:
        max_number_of_items = int(
            et_element.find("MAX-NUMBER-OF-ITEMS").text)
    else:
        max_number_of_items = None

    is_visible = et_element.get("IS-VISIBLE")
    if is_visible is not None:
        assert is_visible in ["true", "false"]
        is_visible = is_visible == "true"
    else:
        is_visible = False
    eopf = EndOfPduField(id,
                         short_name,
                         long_name=long_name,
                         description=description,
                         structure_ref=structure_ref,
                         structure_snref=structure_snref,
                         min_number_of_items=min_number_of_items,
                         max_number_of_items=max_number_of_items,
                         is_visible=is_visible)

    return eopf



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import List
from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..diagcodedtypes import DiagCodedType
from ..odxtypes import DataType
from ..exceptions import DecodeError, EncodeError

from .parameterbase import Parameter


class NrcConstParameter(Parameter):
    """A param of type NRC-CONST defines a set of values to be matched.

    An NRC-CONST can only be used in a negative response.
    Its encoding behaviour is similar to a VALUE parameter with a TEXTTABLE.
    However, an NRC-CONST is used for matching a response (similar to a CODED-CONST).

    See ASAM MCD-2 D (ODX), p. 77-79.
    """

    def __init__(self, short_name, diag_coded_type: DiagCodedType, coded_values: List[int], **kwargs):
        super().__init__(short_name,
                         parameter_type="NRC-CONST", **kwargs)

        self._diag_coded_type = diag_coded_type
        # TODO: Does it have to be an integer or is that just common practice?
        assert all(isinstance(coded_value, int)
                   for coded_value in coded_values)
        self.coded_values = coded_values

    @property
    def diag_coded_type(self):
        return self._diag_coded_type

    @property
    def bit_length(self):
        return self.diag_coded_type.bit_length

    @property
    def internal_data_type(self) -> DataType:
        return self.diag_coded_type.base_data_type

    def is_required(self):
        return False

    def is_optional(self):
        return False

    def get_coded_value(self):
        return self.coded_value

    def get_coded_value_as_bytes(self, encode_state: EncodeState):
        if self.short_name in encode_state.parameter_values:
            if encode_state.parameter_values[self.short_name] not in self.coded_values:
                raise EncodeError(f"The parameter '{self.short_name}' must have"
                                  f" one of the constant values {self.coded_values}")
            else:
                coded_value = encode_state.parameter_values[self.short_name]
        else:
            # If the user does not select one, just select any.
            # I think it does not matter ...
            coded_value = self.coded_values[0]
        return self.diag_coded_type.convert_internal_to_bytes(coded_value, encode_state, bit_position=self.bit_position)

    def decode_from_pdu(self, decode_state: DecodeState):
        if self.byte_position is not None and self.byte_position != decode_state.next_byte_position:
            # Update byte position
            decode_state = decode_state._replace(
                next_byte_position=self.byte_position)

        # Extract coded values
        coded_value, next_byte_position = self.diag_coded_type.convert_bytes_to_internal(decode_state,
                                                                                         bit_position=self.bit_position)

        # Check if the coded value in the message is correct.
        if coded_value not in self.coded_values:
            raise DecodeError(
                f"Coded constant parameter does not match! "
                f"The parameter {self.short_name} expected a coded value in {self.coded_values} but got {coded_value} "
                f"at byte position {decode_state.next_byte_position} "
                f"in coded message {decode_state.coded_message.hex()}."
            )

        return coded_value, next_byte_position

    def _as_dict(self):
        d = super()._as_dict()
        if self.bit_length is not None:
            d["bit_length"] = self.bit_length
        d["coded_values"] = self.coded_values
        return d

    def __repr__(self):
        repr_str = f"CodedConstParameter(short_name='{self.short_name}', coded_values={self.coded_values}"
        if self.long_name is not None:
            repr_str += f", long_name='{self.long_name}'"
        if self.byte_position is not None:
            repr_str += f", byte_position='{self.byte_position}'"
        if self.bit_position:
            repr_str += f", bit_position='{self.bit_position}'"
        if self.semantic is not None:
            repr_str += f", semantic='{self.semantic}'"
        repr_str += f", diag_coded_type={repr(self.diag_coded_type)}"
        if self.description is not None:
            repr_str += f", description='{' '.join(self.description.split())}'"
        return repr_str + ")"

    def __str__(self):
        lines = [
            super().__str__(),
        ]
        return "\n".join(lines)

    def get_description_of_valid_values(self) -> str:
        """return a human-understandable description of valid physical values"""
        return f"One of the constant internal values: {self.coded_values}"



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from typing import Optional, Union

from ..dataobjectproperty import DataObjectProperty, DopBase, DtcDop
from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..physicaltype import PhysicalType
from ..globals import logger

from .parameterbase import Parameter


class ParameterWithDOP(Parameter):
    def __init__(self,
                 short_name: str,
                 parameter_type: str,
                 dop=None,
                 dop_ref: Optional[str] = None,
                 dop_snref: Optional[str] = None,
                 **kwargs):
        super().__init__(short_name, parameter_type, **kwargs)
        self.dop_ref = dop_ref
        self.dop_snref = dop_snref

        self._dop = dop
        if dop and not dop_ref and not dop_snref:
            self.dop_ref = dop.id
        assert self.dop_ref or self.dop_snref

    @property
    def dop(self) -> Union[DopBase, None]:
        """may be a DataObjectProperty, a Structure or None"""
        return self._dop

    @dop.setter
    def dop(self, dop: DopBase):
        self._dop = dop
        if not self.dop_snref:
            self.dop_ref = dop.id
        else:
            self.dop_snref = dop.short_name

    def resolve_references(self, parent_dl, id_lookup):
        dop = self.dop
        if self.dop_snref:
            dop = parent_dl.data_object_properties[self.dop_snref]
            if not dop:
                logger.info(
                    f"Param {self.short_name} could not resolve DOP-SNREF {self.dop_snref}")
        elif self.dop_ref:
            dop = id_lookup.get(self.dop_ref)
            if not dop:
                logger.info(
                    f"Param {self.short_name} could not resolve DOP-REF {self.dop_ref}")
        else:
            logger.warn(
                f"Param {self.short_name} without DOP-(SN)REF should not exist!")
        if self.dop and dop != self.dop:
            logger.warn(
                f"Param {self.short_name}: Reference and DOP are inconsistent!")

        self._dop = dop

    @property
    def bit_length(self):
        if self.dop is not None:
            return self.dop.bit_length
        else:
            return None

    @property
    def physical_type(self) -> Optional[PhysicalType]:
        if isinstance(self.dop, (DataObjectProperty, DtcDop)):
            return self.dop.physical_type
        else:
            return None

    def get_coded_value(self, physical_value=None):
        return self.dop.convert_physical_to_internal(physical_value)

    def get_coded_value_as_bytes(self, encode_state: EncodeState):
        assert self.dop is not None, "Reference to DOP is not resolved"
        physical_value = encode_state.parameter_values[self.short_name]
        return self.dop.convert_physical_to_bytes(physical_value, encode_state, bit_position=self.bit_position)

    def decode_from_pdu(self, decode_state: DecodeState):
        assert self.dop is not None, "Reference to DOP is not resolved"
        if self.byte_position is not None and self.byte_position != decode_state.next_byte_position:
            decode_state = decode_state._replace(
                next_byte_position=self.byte_position)

        # Use DOP to decode
        phys_val, next_byte_position = self.dop.convert_bytes_to_physical(
            decode_state, bit_position=self.bit_position)

        return phys_val, next_byte_position

    def _as_dict(self):
        d = super()._as_dict()
        if self.dop is not None:
            if self.bit_length is not None:
                d["bit_length"] = self.bit_length
            d["dop_ref"] = self.dop.id
        elif self.dop_ref is not None:
            d["dop_ref"] = self.dop_ref
        elif self.dop_snref is not None:
            d["dop_snref"] = self.dop_snref

        return d

    def __str__(self):
        lines = [
            super().__str__(),
            " " + str(self.dop).replace("\n", "\n ")
        ]
        return "\n".join(lines)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import zipfile
import os
import odxtools
import jinja2
import time
import datetime
import inspect

from typing import Any, Dict, List, Tuple

def jinja2_odxraise_helper(msg):
    raise Exception(msg)


__module_filname = inspect.getsourcefile(odxtools)
assert isinstance(__module_filname, str)
__pdx_stub_dir = os.path.sep.join([os.path.dirname(__module_filname),
                                  "pdx_stub"])
def write_pdx_file(output_file_name : str,
                   database : odxtools.Database,
                   auxiliary_content_specifiers : List[Tuple[str, str]] = [],
                   stub_dir : str = __pdx_stub_dir) -> bool:
    """
    Write an internalized database to a PDX file.
    """

    file_index = list()

    with zipfile.ZipFile(output_file_name,
                         mode="w",
                         compression=zipfile.ZIP_DEFLATED) as zf:

        # write all files in the stub directory
        for root, dir, files in os.walk(stub_dir):
            for stub_file_name in files:
                # we are not interested in the autosave garbage of
                # editors...
                if stub_file_name.startswith("."):
                    continue
                elif stub_file_name.startswith("#"):
                    continue
                elif stub_file_name.endswith("~"):
                    continue
                elif stub_file_name.endswith(".bak"):
                    continue
                elif stub_file_name.endswith(".tpl"):
                    continue

                stub_file_mime_type = "text/plain"
                if stub_file_name.endswith(".odx-cs"):
                    stub_file_mime_type = "application/x-asam.odx.odx-cs"
                elif stub_file_name.endswith(".odx-d"):
                    stub_file_mime_type = "application/x-asam.odx.odx-d"

                in_path = list([root])
                in_path.append(stub_file_name)
                in_file_name = os.path.sep.join(in_path)

                stub_file_stats = os.stat(in_file_name)
                stub_file_cdate = datetime.datetime.fromtimestamp(stub_file_stats.st_ctime)
                stub_file_creation_date = stub_file_cdate.strftime("%Y-%m-%dT%H:%M:%S")
                file_index.append( (stub_file_name,
                                    stub_file_creation_date,
                                    stub_file_mime_type) )
                with zf.open(stub_file_name, "w") as out_file:
                    out_file.write(open(in_file_name, "rb").read())

        # write the auxiliary files
        for output_file_name, data in auxiliary_content_specifiers:
            file_cdate = datetime.datetime.fromtimestamp(time.time())
            creation_date = file_cdate.strftime("%Y-%m-%dT%H:%M:%S")

            mime_type = "text/plain"
            if stub_file_name.endswith(".odx-cs"):
                mime_type = "application/x-asam.odx.odx-cs"
            elif stub_file_name.endswith(".odx-d"):
                mime_type = "application/x-asam.odx.odx-d"

            zf_name = os.path.basename(output_file_name)
            with zf.open(zf_name, "w") as out_file:
                file_index.append( (zf_name, creation_date, mime_type) )
                out_file.write(data)  # type: ignore

        jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(stub_dir))
        jinja_env.globals['odxraise'] = jinja2_odxraise_helper

        # allows to put XML attributes on a separate line while it is
        # collapsed with the previous line in the rendering
        jinja_env.filters["odxtools_collapse_xml_attribute"] = lambda x: " "+x.strip() if x.strip() else ""
        
        vars: Dict[str, Any] = {}
        vars["odxtools_version"] = odxtools.__version__
        vars["database"] = database
        vars["dlc"] = None

        # write the actual diagnostic data.
        dlc_tpl = jinja_env.get_template("diag_layer_container.odx-d.tpl")
        for dlc in database.diag_layer_containers:
            vars["dlc"] = dlc

            file_name = f"{dlc.short_name}.odx-d"
            file_cdate = datetime.datetime.fromtimestamp(time.time())
            creation_date = file_cdate.strftime("%Y-%m-%dT%H:%M:%S")
            mime_type = "application/x-asam.odx.odx-d"

            file_index.append( (file_name, creation_date, mime_type) )
            zf.writestr(file_name, dlc_tpl.render(**vars))

        del vars["dlc"]

        # write the index.xml file
        vars["file_index"] = file_index
        index_tpl = jinja_env.get_template("index.xml.tpl")
        text = index_tpl.render(**vars)
        zf.writestr("index.xml", text)

    return True



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..diagcodedtypes import DiagCodedType
from ..odxtypes import DataType
from ..exceptions import DecodeError

from .parameterbase import Parameter


class CodedConstParameter(Parameter):
    def __init__(self, short_name, diag_coded_type: DiagCodedType, coded_value, **kwargs):
        super().__init__(short_name,
                         parameter_type="CODED-CONST", **kwargs)

        self._diag_coded_type = diag_coded_type
        assert isinstance(coded_value, int)
        self.coded_value = coded_value

    @property
    def diag_coded_type(self):
        return self._diag_coded_type

    @property
    def bit_length(self):
        return self.diag_coded_type.bit_length

    @property
    def internal_data_type(self) -> DataType:
        return self.diag_coded_type.base_data_type

    def is_required(self):
        return False

    def is_optional(self):
        return False

    def get_coded_value(self):
        return self.coded_value

    def get_coded_value_as_bytes(self, encode_state: EncodeState):
        if self.short_name in encode_state.parameter_values \
                and encode_state.parameter_values[self.short_name] != self.coded_value:
            raise TypeError(f"The parameter '{self.short_name}' is constant {self.coded_value}"
                            " and thus can not be changed.")
        return self.diag_coded_type.convert_internal_to_bytes(self.coded_value, encode_state, bit_position=self.bit_position)

    def decode_from_pdu(self, decode_state: DecodeState):
        if self.byte_position is not None and self.byte_position != decode_state.next_byte_position:
            # Update byte position
            decode_state = decode_state._replace(
                next_byte_position=self.byte_position)

        # Extract coded values
        coded_val, next_byte_position = self.diag_coded_type.convert_bytes_to_internal(decode_state,
                                                                                       bit_position=self.bit_position)

        # Check if the coded value in the message is correct.
        if self.coded_value != coded_val:
            raise DecodeError(
                f"Coded constant parameter does not match! "
                f"The parameter {self.short_name} expected coded value {self.coded_value} but got {coded_val} "
                f"at byte position {decode_state.next_byte_position} "
                f"in coded message {decode_state.coded_message.hex()}."
            )

        return self.coded_value, next_byte_position

    def _as_dict(self):
        d = super()._as_dict()
        if self.bit_length is not None:
            d["bit_length"] = self.bit_length
        d["coded_value"] = hex(self.coded_value)
        return d

    def __repr__(self):
        repr_str = f"CodedConstParameter(short_name='{self.short_name}', coded_value={self.coded_value}"
        if self.long_name is not None:
            repr_str += f", long_name='{self.long_name}'"
        if self.byte_position is not None:
            repr_str += f", byte_position='{self.byte_position}'"
        if self.bit_position:
            repr_str += f", bit_position='{self.bit_position}'"
        if self.semantic is not None:
            repr_str += f", semantic='{self.semantic}'"
        repr_str += f", diag_coded_type={repr(self.diag_coded_type)}"
        if self.description is not None:
            repr_str += f", description='{' '.join(self.description.split())}'"
        return repr_str + ")"

    def __str__(self):
        lines = [
            super().__str__(),
        ]
        return "\n".join(lines)

    def get_description_of_valid_values(self) -> str:
        """return a human-understandable description of valid physical values"""
        return f"Constant internal value: {self.coded_value}"



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from .nameditemlist import NamedItemList
from .utils import read_description_from_odx

from dataclasses import dataclass, field
from typing import Optional, List

@dataclass()
class XDoc:
    short_name : str
    long_name : Optional[str] = None
    description : Optional[str] = None
    number: Optional[str] = None
    state: Optional[str] = None
    date: Optional[str] = None
    publisher: Optional[str] = None
    url: Optional[str] = None
    position: Optional[str] = None

@dataclass()
class RelatedDoc:
    description: Optional[str] = None
    xdoc: Optional[XDoc] = None

@dataclass()
class CompanySpecificInfo:
    related_docs: Optional[List[RelatedDoc]]

@dataclass()
class TeamMember:
    id: str
    short_name: str
    long_name: Optional[str] = None
    description: Optional[str] = None
    roles: Optional[List[str]] = None
    department: Optional[str] = None
    address: Optional[str] = None
    zip: Optional[str] = None
    city: Optional[str] = None
    phone: Optional[str] = None
    fax: Optional[str] = None
    email: Optional[str] = None

@dataclass()
class CompanyData:
    id: str
    short_name: str
    long_name: Optional[str] = None
    description: Optional[str] = None
    roles: Optional[List[str]] = None
    team_members: Optional[NamedItemList[TeamMember]] = None
    company_specific_info: Optional[CompanySpecificInfo] = None

    def _build_id_lookup(self):
        result = { self.id: self }

        # team members
        if self.team_members is not None:
            for tm in self.team_members:
                result[tm.id] = tm

        return result

    def _resolve_references(self, id_lookup):
        pass

def read_xdoc_from_odx(xdoc):
    short_name = xdoc.find("SHORT-NAME").text

    long_name = xdoc.find("LONG-NAME")
    if long_name is not None:
        long_name = long_name.text

    description = xdoc.find("DESC")
    if description is not None:
        description = read_description_from_odx(description)

    number = xdoc.find("NUMBER")
    if number is not None:
        number = number.text

    state = xdoc.find("STATE")
    if state is not None:
        state = state.text

    date = xdoc.find("DATE")
    if date is not None:
        date = date.text

    publisher = xdoc.find("PUBLISHER")
    if publisher is not None:
        publisher = publisher.text

    url = xdoc.find("URL")
    if url is not None:
        url = url.text

    position = xdoc.find("POSITION")
    if position is not None:
        position = position.text

    return XDoc(short_name=short_name,
                long_name=long_name,
                description=description,
                number=number,
                state=state,
                date=date,
                publisher=publisher,
                url=url,
                position=position,
                )

def read_company_datas_from_odx(et_element):
    if et_element is None:
        return None

    cdl = NamedItemList(lambda x: x.short_name)

    for cd in et_element.iterfind("COMPANY-DATA"):
        id = cd.attrib["ID"]
        short_name = cd.find("SHORT-NAME").text

        long_name = cd.find("LONG-NAME")
        if long_name is not None:
            long_name = long_name.text

        description = cd.find("DESC")
        if description is not None:
            description = read_description_from_odx(description)

        roles = cd.find("ROLES")
        if roles is not None:
            rlist = list()

            for role in roles.iterfind("ROLE"):
                rlist.append(role.text)

            roles = rlist

        team_members = cd.find("TEAM-MEMBERS")
        if team_members is not None:
            tml = NamedItemList(lambda x: x.short_name)

            for tm in team_members.iterfind("TEAM-MEMBER"):
                tm_id = tm.attrib["ID"]
                tm_short_name = tm.find("SHORT-NAME").text

                tm_long_name = tm.find("LONG-NAME")
                if tm_long_name is not None:
                    tm_long_name = tm_long_name.text

                tm_description = tm.find("DESC")
                if tm_description is not None:
                    tm_description = read_description_from_odx(tm_description)

                tm_roles = tm.find("ROLES")
                if tm_roles is not None:
                    rlist = list()
                    for role in tm_roles.iterfind("ROLE"):
                        rlist.append(role.text)
                    tm_roles = rlist

                tm_department = tm.find("DEPARTMENT")
                if tm_department is not None:
                    tm_department = tm_department.text

                tm_address = tm.find("ADDRESS")
                if tm_address is not None:
                    tm_address = tm_address.text

                tm_zip = tm.find("ZIP")
                if tm_zip is not None:
                    tm_zip = tm_zip.text

                tm_city = tm.find("CITY")
                if tm_city is not None:
                    tm_city = tm_city.text

                tm_phone = tm.find("PHONE")
                if tm_phone is not None:
                    tm_phone = tm_phone.text

                tm_fax = tm.find("FAX")
                if tm_fax is not None:
                    tm_fax = tm_fax.text

                tm_email = tm.find("EMAIL")
                if tm_email is not None:
                    tm_email = tm_email.text

                tml.append(TeamMember(id=tm_id,
                                      short_name=tm_short_name,
                                      long_name=tm_long_name,
                                      description=tm_description,
                                      roles=tm_roles,
                                      department=tm_department,
                                      address=tm_address,
                                      zip=tm_zip,
                                      city=tm_city,
                                      phone=tm_phone,
                                      fax=tm_fax,
                                      email=tm_email))

            team_members = tml

        company_specific_info = cd.find("COMPANY-SPECIFIC-INFO")
        if company_specific_info is not None:
            related_docs = company_specific_info.find("RELATED-DOCS")
            if related_docs is not None:
                rdlist = list()
                for rd in related_docs.iterfind("RELATED-DOC"):
                    rd_description = rd.find("DESC")
                    if rd_description is not None:
                        rd_description = read_description_from_odx(rd_description)

                    xdoc = rd.find("XDOC")
                    if xdoc is not None:
                        xdoc = read_xdoc_from_odx(xdoc)

                    rdlist.append(RelatedDoc(description=rd_description,
                                             xdoc=xdoc,
                                             ))

                related_docs = rdlist

            company_specific_info = CompanySpecificInfo(related_docs=related_docs)

        cdl.append(CompanyData(id=id,
                               short_name=short_name,
                               long_name=long_name,
                               description=description,
                               roles=roles,
                               team_members=team_members,
                               company_specific_info=company_specific_info))

    return cdl



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from typing import Any, Dict, List, Optional, Union, Type

from ..odxtypes import DataType
from ..utils import read_description_from_odx
from ..globals import logger

from .compumethodbase import CompuMethod
from .compuscale import CompuScale
from .identicalcompumethod import IdenticalCompuMethod
from .limit import IntervalType, Limit
from .linearcompumethod import LinearCompuMethod
from .scalelinearcompumethod import ScaleLinearCompuMethod
from .tabintpcompumethod import TabIntpCompuMethod
from .texttablecompumethod import TexttableCompuMethod


def _parse_compu_scale_to_linear_compu_method(scale_element,
                                              internal_type: DataType,
                                              physical_type: DataType,
                                              is_scale_linear=False,
                                              additional_kwargs={}):
    assert physical_type in [DataType.A_FLOAT32,
                             DataType.A_FLOAT64,
                             DataType.A_INT32,
                             DataType.A_UINT32]
    assert internal_type in [DataType.A_FLOAT32,
                             DataType.A_FLOAT64,
                             DataType.A_INT32,
                             DataType.A_UINT32]

    computation_python_type: Union[Type[float], Type[int]]
    if internal_type.as_python_type() == float or physical_type.as_python_type() == float:
        computation_python_type = float
    else:
        computation_python_type = int

    kwargs = additional_kwargs.copy()
    kwargs["internal_type"] = internal_type
    kwargs["physical_type"] = physical_type

    coeffs = scale_element.find("COMPU-RATIONAL-COEFFS")
    nums = coeffs.iterfind("COMPU-NUMERATOR/V")

    offset = computation_python_type(next(nums).text)
    factor = computation_python_type(next(nums).text)
    if coeffs.find("COMPU-DENOMINATOR/V") is not None:
        kwargs["denominator"] = float(
            coeffs.find("COMPU-DENOMINATOR/V").text)
        assert kwargs["denominator"] > 0

    # Read lower limit
    internal_lower_limit = read_limit_from_odx(
        scale_element.find("LOWER-LIMIT"),
        internal_type=internal_type
    )
    if internal_lower_limit is None:
        internal_lower_limit = Limit(float("-inf"), IntervalType.INFINITE)
    kwargs["internal_lower_limit"] = internal_lower_limit

    # Read upper limit
    internal_upper_limit = read_limit_from_odx(
        scale_element.find("UPPER-LIMIT"),
        internal_type=internal_type
    )
    if internal_upper_limit is None:
        if not is_scale_linear:
            internal_upper_limit = Limit(float("inf"), IntervalType.INFINITE)
        else:
            assert (internal_lower_limit is not None
                    and internal_lower_limit.interval_type == IntervalType.CLOSED)
            logger.info("Scale linear without UPPER-LIMIT")
            internal_upper_limit = internal_lower_limit
    kwargs["internal_upper_limit"] = internal_upper_limit

    return LinearCompuMethod(offset=offset, factor=factor, **kwargs)


def read_limit_from_odx(et_element, internal_type: DataType):
    limit: Optional[Limit] = None
    if et_element is not None:
        if et_element.get("INTERVAL-TYPE"):
            interval_type = IntervalType(et_element.get("INTERVAL-TYPE"))
        else:
            interval_type = IntervalType.CLOSED

        if interval_type == IntervalType.INFINITE:
            if et_element.tag == "LOWER-LIMIT":
                limit = Limit(float("-inf"), interval_type)
            else:
                assert et_element.tag == "UPPER-LIMIT"
                limit = Limit(float("inf"), interval_type)
        else:
            if internal_type == DataType.A_BYTEFIELD:
                limit = Limit(int(et_element.text, 16), interval_type)
            else:
                limit = Limit(internal_type.from_string(et_element.text),
                              interval_type)
    return limit


def read_compu_method_from_odx(et_element, internal_type: DataType, physical_type: DataType) -> CompuMethod:
    compu_category = et_element.find("CATEGORY").text
    assert compu_category in ["IDENTICAL", "LINEAR", "SCALE-LINEAR",
                              "TEXTTABLE", "COMPUCODE", "TAB-INTP",
                              "RAT-FUNC", "SCALE-RAT-FUNC"]

    if et_element.find("COMPU-PHYS-TO-INTERNAL") is not None:  # TODO: Is this never used?
        raise NotImplementedError(
            f"Found COMPU-PHYS-TO-INTERNAL for category {compu_category}")

    kwargs: Dict[str, Any] = {"internal_type": internal_type}

    if compu_category == "IDENTICAL":
        assert (internal_type == physical_type or (
            internal_type in [DataType.A_ASCIISTRING,  DataType.A_UTF8STRING] and physical_type == DataType.A_UNICODE2STRING)
        ), (f"Internal type '{internal_type}' and physical type '{physical_type}'"
            f" must be the same for compu methods of category '{compu_category}'")
        return IdenticalCompuMethod(internal_type=internal_type, physical_type=physical_type)

    if compu_category == "TEXTTABLE":
        assert physical_type == DataType.A_UNICODE2STRING
        compu_internal_to_phys = et_element.find("COMPU-INTERNAL-TO-PHYS")

        internal_to_phys: List[CompuScale] = []
        for scale in compu_internal_to_phys.iterfind("COMPU-SCALES/COMPU-SCALE"):
            lower_limit = read_limit_from_odx(scale.find("LOWER-LIMIT"),
                                              internal_type=internal_type)
            upper_limit = read_limit_from_odx(scale.find("UPPER-LIMIT"),
                                              internal_type=internal_type)

            if (vt := scale.find("COMPU-INVERSE-VALUE/VT")) is not None:
                compu_inverse_value = internal_type.from_string(vt)
            elif (v := scale.find("COMPU-INVERSE-VALUE/V")) is not None:
                compu_inverse_value = internal_type.from_string(v)
            else:
                compu_inverse_value = None

            internal_to_phys.append(CompuScale(
                short_label=(scale.find("SHORT-LABEL").text
                             if scale.find("SHORT-LABEL") is not None else None),
                description=read_description_from_odx(scale.find("DESC")),
                lower_limit=lower_limit,
                upper_limit=upper_limit,
                compu_inverse_value=compu_inverse_value,
                compu_const=scale.find("COMPU-CONST").find("VT").text
            ))

        kwargs["internal_to_phys"] = internal_to_phys
        return TexttableCompuMethod(**kwargs)

    elif compu_category == "LINEAR":
        # Compu method can be described by the function f(x) = (offset + factor * x) / denominator

        scale = et_element.find(
            "COMPU-INTERNAL-TO-PHYS/COMPU-SCALES/COMPU-SCALE")
        return _parse_compu_scale_to_linear_compu_method(scale, internal_type, physical_type, additional_kwargs=kwargs)

    elif compu_category == "SCALE-LINEAR":

        scales = et_element.iterfind(
            "COMPU-INTERNAL-TO-PHYS/COMPU-SCALES/COMPU-SCALE")
        linear_methods = [_parse_compu_scale_to_linear_compu_method(
            scale, internal_type, physical_type, additional_kwargs=kwargs) for scale in scales]
        return ScaleLinearCompuMethod(linear_methods)

    elif compu_category == "TAB-INTP":
        return TabIntpCompuMethod(internal_type=internal_type, physical_type=physical_type)

    # TODO: Implement other categories (never instantiate CompuMethod)
    logger.warning(
        f"Warning: Computation category {compu_category} is not implemented!")
    return CompuMethod(DataType.A_UINT32, DataType.A_UINT32, f"NOT-IMPLEMENTED:{compu_category}")



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from ..decodestate import DecodeState
from ..encodestate import EncodeState

from .parameterbase import Parameter


class MatchingRequestParameter(Parameter):
    def __init__(self,
                 short_name,
                 request_byte_position,
                 byte_length,
                 **kwargs):
        super().__init__(short_name,
                         parameter_type="MATCHING-REQUEST-PARAM",
                         **kwargs)
        assert byte_length is not None
        assert request_byte_position is not None
        self.request_byte_position = request_byte_position
        self._byte_length = byte_length

    @property
    def bit_length(self):
        return 8 * self._byte_length

    @property
    def byte_length(self):
        return self._byte_length

    def is_required(self):
        return True

    def is_optional(self):
        return False

    def get_coded_value(self, request_value=None):
        return request_value

    def get_coded_value_as_bytes(self, encode_state: EncodeState):
        if not encode_state.triggering_request:
            raise TypeError(f"Parameter '{self.short_name}' is of matching request type,"
                            " but no original request has been specified.")
        return encode_state.triggering_request[self.request_byte_position:self.request_byte_position+self.byte_length]

    def decode_from_pdu(self, decode_state: DecodeState):
        byte_position = self.byte_position if self.byte_position is not None else decode_state.next_byte_position
        byte_length = (self.bit_length + self.bit_position + 7) // 8
        val_as_bytes = decode_state.coded_message[byte_position:byte_position+byte_length]

        return val_as_bytes, byte_position + byte_length

    def __repr__(self):
        repr_str = f"MatchingRequestParameter(short_name='{self.short_name}'"
        if self.long_name is not None:
            repr_str += f", long_name='{self.long_name}'"
        if self.byte_position is not None:
            repr_str += f", byte_position='{self.byte_position}'"
        if self.bit_position:
            repr_str += f", bit_position='{self.bit_position}'"
        if self.semantic is not None:
            repr_str += f", semantic='{self.semantic}'"
        if self.request_byte_position is not None:
            repr_str += f", request_byte_position='{self.request_byte_position}'"
        if self.byte_length is not None:
            repr_str += f", byte_length='{self.byte_length}'"
        if self.description is not None:
            repr_str += f", description='{' '.join(self.description.split())}'"
        return repr_str + ")"

    def __str__(self):
        return super().__str__() + f"\n Request byte position = {self.request_byte_position}, byte length = {self._byte_length}"



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass
from typing import Optional


@dataclass()
class StateTransition:
    """
    Corresponds to STATE.
    """
    id: str
    short_name: str
    long_name: Optional[str] = None
    source_short_name: Optional[str] = None
    target_short_name: Optional[str] = None


def read_state_transition_from_odx(et_element):
    short_name = et_element.find("SHORT-NAME").text
    id = et_element.get("ID")

    long_name = et_element.find(
        "LONG-NAME").text if et_element.find("LONG-NAME") is not None else None
    source_short_name = et_element.find("SOURCE-SNREF").attrib["SHORT-NAME"] if et_element.find("SOURCE-SNREF") is not None else None
    target_short_name = et_element.find("TARGET-SNREF").attrib["SHORT-NAME"] if et_element.find("TARGET-SNREF") is not None else None

    return StateTransition(id=id,
                           short_name=short_name,
                           long_name=long_name,
                           source_short_name=source_short_name,
                           target_short_name=target_short_name)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from typing import List, NamedTuple, Optional


class CompuRationalCoeffs(NamedTuple):
    numerators: List[float]
    denominators: Optional[List[float]] = None



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from enum import IntEnum

class SID(IntEnum):
    """The standardized service IDs for ODB-II (SAE J1979).

    The physical CAN address to which OBD-II (functional) requests are
    to be send to is usually 0x7DF and CAN frames must be padded to 8
    bytes (with 0xcc as the padding pattern, i.e., one byte by ISO-TP
    and "actual" 7 payload bytes). Some vehicles support addressing up
    to seven additional ECUs via OBD-II which use the CAN IDs 0x7DF +
    $I.

    The physical CAN ID for responses of such requests is the CAN
    request ID plus 8 (i.e. usually 0x7E8).

    For additional information, see https://en.wikipedia.org/wiki/OBD-II_PIDs

    """

    # 0x00..0x0a: service IDs for emissions systems from ISO15031-5
    ShowCurrentData = 0x01
    ShowFreezeFrameData = 0x02
    ShowStoredDiagnosticTroubleCodes = 0x03
    ClearDiagnosticTroubleCodesAndStoredValues = 0x04
    OxygenSensorMonitoringTestResults = 0x05
    OtherTestResults = 0x06
    ShowPendingDiagnosticTroubleCodes = 0x07
    ControlOperationOfOnboardComponentOrSystem = 0x08
    RequestVehicleInformation = 0x09
    PermanentDiagnosticTroubleCodes = 0x0A



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass
from enum import Enum
from typing import Optional

from .odxtypes import DataType


class Radix(Enum):
    HEX = "HEX"
    DEC = "DEC"
    BIN = "BIN"
    OCT = "OCT"


@dataclass
class PhysicalType:
    """The physical type describes the base data type of a parameter.

    Similar to how a `DiagCodedType` describes the encoding of the internal value,
    the `PhysicalType` describes how to display the physical value.

    For an unsigned integers (A_UINT32) it may specify a display radix (HEX, DEC, BIN, OCT).
    For floating point numbers (A_FLOAT32, A_FLOAT64) it may specify a precision,
    that is, the number of digits to display after the decimal point.

    Examples
    --------

    A hexadecimal, integer number::

    PhysicalType(DataType.A_UINT32, display_radix=Radix.HEX)

    A floating point number that should be displayed with 2 digits after the decimal point::

    PhysicalType(DataType.A_FLOAT64, precision=2)
    """

    base_data_type: DataType

    display_radix: Optional[Radix] = None
    """The display radix defines how integers are displayed to the user.
    The display radix is only applicable if the base data type is A_UINT32. 
    """

    precision: Optional[int] = None
    """Number of digits after the decimal point to display to the user
    The precision is only applicable if the base data type is A_FLOAT32 or A_FLOAT64. 
    """

    def __post_init__(self):
        self.base_data_type = DataType(self.base_data_type)
        if self.display_radix is not None:
            self.display_radix = Radix(self.display_radix)


def read_physical_type_from_odx(et_element):
    base_data_type = et_element.get("BASE-DATA-TYPE")
    assert base_data_type in ["A_INT32", "A_UINT32", "A_FLOAT32", "A_FLOAT64",
                              "A_ASCIISTRING", "A_UTF8STRING", "A_UNICODE2STRING", "A_BYTEFIELD"]
    display_radix = et_element.get("DISPLAY-RADIX")
    precision = et_element.findtext("PRECISION")
    if precision is not None:
        precision = int(precision)

    return PhysicalType(base_data_type,
                        display_radix=display_radix,
                        precision=precision)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import argparse
from typing import List, Union

from ._print_utils import print_diagnostic_service, format_desc
from . import _parser_utils

from ..database import Database
from ..service import DiagService
from ..singleecujob import SingleEcuJob


def print_summary(odxdb: Database,
                  print_services=False,
                  print_dops=False,
                  print_params=False,
                  print_com_params=False,
                  print_pre_condition_states=False,
                  print_state_transitions=False,
                  print_audiences=False,
                  allow_unknown_bit_lengths=False,
                  variants=None,
                  service_filter=lambda x: True):

    diag_layer_names = variants if variants else [
        dl.short_name for dl in odxdb.diag_layers
    ]

    for dl_sn in diag_layer_names:
        dl = odxdb.diag_layers[dl_sn]
        if not dl:
            print(f"The variant '{dl_sn}' could not be found!")
            continue

        all_services: List[Union[DiagService, SingleEcuJob]] \
            = sorted(dl.services, key=lambda x: x.short_name)

        data_object_properties = dl.data_object_properties
        com_params = dl.communication_parameters

        if dl.get_receive_id() is not None:
            recv_id = hex(dl.get_receive_id())
        else:
            recv_id = "None"
        if dl.get_send_id() is not None:
            send_id = hex(dl.get_send_id())
        else:
            send_id = "None"
        print(
            f"{dl.variant_type} '{dl.short_name}' (Receive ID: {recv_id}, Send ID: {send_id})"
        )
        print(
            f" num services: {len(all_services)}, num DOPs: {len(data_object_properties)}, num communication parameters: {len(com_params)}."
        )

        if dl.description:
            desc = format_desc(dl.description, ident=2)
            print(f" Description: " + desc)

        if print_services and len(all_services) > 0:
            services = [s for s in all_services if service_filter(s)]
            if len(services) > 0:
                print(
                    f"The services of the {dl.variant_type} '{dl.short_name}' are: ")
                for service in services:
                    if isinstance(service, DiagService):
                        print_diagnostic_service(
                            service,
                            print_params=print_params,
                            print_pre_condition_states=print_pre_condition_states,
                            print_state_transitions=print_state_transitions,
                            print_audiences=print_audiences,
                            allow_unknown_bit_lengths=allow_unknown_bit_lengths
                        )
                    elif isinstance(service, SingleEcuJob):
                        print(f" Single ECU job: {service.id}")
                    else:
                        print(f" Unidentifiable service: {service}")

        if print_dops and len(data_object_properties) > 0:
            print(f"The DOPs of the {dl.variant_type} '{dl.short_name}' are: ")
            for dop in sorted(data_object_properties, key=lambda x: (type(x).__name__, x.short_name)):
                print("  " + str(dop).replace("\n", "\n  "))

        if print_com_params and len(com_params) > 0:
            print(
                f"The communication parameters of the {dl.variant_type} '{dl.short_name}' are: ")
            for com_param in com_params:
                print(f"  {com_param.id_ref}: {com_param.value}")


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        "list",
        description="\n".join([
            "List the content of automotive diagnostic files (*.pdx)",
            "",
            "Examples:",
            "  For displaying only the names of the diagnostic layers use:",
            "    odxtools list ./path/to/database.pdx",
            "  For displaying all content use:",
            "    odxtools list ./path/to/database.pdx --all",
            "  For more information use:",
            "    odxtools list -h"
        ]),
        help="Print a summary of automotive diagnostic files.",
        formatter_class=argparse.RawTextHelpFormatter)
    _parser_utils.add_pdx_argument(parser)

    parser.add_argument("-v", "--variants", nargs='+', metavar="VARIANT",
                        required=False, help="Specifies which variants should be included.", default="all")

    # The service option is None if option is not passed at all (-> do not print services). It is an empty list if --services is passed
    parser.add_argument("-s", "--services", nargs='*', default=None, metavar="SERVICE",
                        required=False, help="Print a list of diagnostic services specified in the pdx. \n"
                        + "If no service names are specified, all services are printed.")
    # Pretty print message format and list parameters
    parser.add_argument("-p", "--params", default=False, action='store_const', const=True,
                        required=False, help="Print a list of all parameters relevant for the selected items.\n")
    parser.add_argument("-d", "--dops", default=False, action='store_const', const=True,
                        required=False, help="Print a list of all data object properties relevant for the selected items")

    # Shortcut to just dump everything
    parser.add_argument("-a", "--all", default=False, action='store_const', const=True,
                        required=False, help="Print a list of all diagnostic services and DOPs specified in the pdx")


def run(args):
    odxdb = _parser_utils.load_file(args)

    variants = args.variants if args.variants else None
    print_summary(odxdb, print_services=args.all or args.params or args.services is not None,
                  service_filter=(lambda s: s.short_name in args.services
                                  if args.services and len(args.services) > 0 else lambda s: True),
                  print_dops=args.all or args.dops, variants=None if variants == "all" else variants,
                  print_params=args.all or args.params,
                  print_com_params=args.all,
                  print_pre_condition_states=args.all,
                  print_state_transitions=args.all,
                  print_audiences=args.all,
                  allow_unknown_bit_lengths=args.all)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from enum import IntEnum
from itertools import chain
from typing import Union

import odxtools.obd as obd

class UDSSID(IntEnum):
    """The service IDs standardized by UDS.

    For additional information, see https://en.wikipedia.org/wiki/Unified_Diagnostic_Services
    """
    # 0x10..0x3e: UDS standardized service IDs
    DiagnosticSessionControl = 0x10
    EcuReset = 0x11
    SecurityAccess = 0x27
    CommunicationControl = 0x28
    Authentication = 0x29
    TesterPresent = 0x3E
    AccessTimingParameters = 0x83
    SecuredDataTransmission = 0x84
    ControlDtcSettings = 0x85
    ResponseOnEvent = 0x86
    LinkControl = 0x87
    ReadDataByIdentifier = 0x22
    ReadMemoryByAddress = 0x23
    ReadScalingDataByIdentifier = 0x24
    ReadDataByIdentifierPeriodic = 0x2A
    DynamicallyDefineDataIdentifier = 0x2C
    WriteDataByIdentifier = 0x2E
    WriteMemoryByAddress = 0x3D
    ClearDiagnosticInformation = 0x14
    ReadDtcInformation = 0x19
    InputOutputControlByIdentifier = 0x2F
    RoutineControl = 0x31
    RequestDownload = 0x34
    RequestUpload = 0x35
    TransferData = 0x36
    RequestTransferExit = 0x37
    RequestFileTransfer = 0x38

    # 0x81..0x82: KWP on K-Line

    # 0xA0..0xB9: Reserved for OEM specific services

    # 0xBA..0xBE: Reserved for ECU specific services

# add the OBD SIDs to the ones from UDS
SID = IntEnum('UdsSID', ((i.name, i.value) for i in chain(obd.SID, UDSSID)))  # type: ignore

class NegativeResponseCodes(IntEnum):
    """The standardized negative response codes of UDS.

    In UDS, negative responses always exhibit the format `[0x7f,
    service_id_of_request, response_code]`. The meaning of the last
    byte is what's defined here.
    """
    GeneralReject = 0x10
    ServiceNotSupported = 0x11
    SubFunctionNotSupported = 0x12
    InvalidFormat = 0x13 # Incorrect message length of request or invalid format
    TooLong = 0x14 # Response would be too long
    Busy = 0x21 # please repeat!
    ConditionsIncorrect = 0x22 # request cannot be satisfied because ECU is in wrong state
    RequestSequenceError = 0x24
    NoResponseFromSubNetComponent = 0x25 # we pinged a slave ECU bit it did not respond in time
    Failure = 0x26 # failure prevents execution of requested action
    RequestOutOfRange = 0x31
    SecurityAccessDenied = 0x33
    InvalidKey = 0x35
    ExceededNumberOfAttempts = 0x36
    RequiredTimeDelayNotExpired = 0x37
    Reserved = 0x4F # reserved by Extended Data Link Security Document
    UpDownloadNotAccepted = 0x70
    TransferDataSuspended = 0x71
    GeneralProgrammingFailure = 0x72
    WrongBlockSequenceCounter = 0x73
    ResponsePending = 0x78 # Request correctly received, but response is still pending
    SubFunctionNotSupportedInActiveSession = 0x7E
    ServiceNotSupportedInActiveSession = 0x7F

# get the ID of a positive response
def positive_response_id(service_id: int) -> int:
    """Given a service ID of a request, return the corresponding SID for a positive response"""
    assert service_id != 0x7f - 0x40
    return service_id + 0x40

NegativeResponseId = 0x7f
def negative_response_id(service_id: int) -> int:
    """Given a service ID of a request, return the corresponding SID for a negative response"""
    assert service_id != 0x7f - 0x40  # TODO: What is this assert supposed to do?
    return NegativeResponseId

def is_reponse_pending(telegram_payload: Union[bytes, bytearray], request_sid: int = None) -> bool:
    # "response pending" responses exhibit at least three bytes
    if len(telegram_payload) < 3:
        return False

    sid = telegram_payload[0]
    rq_sid = telegram_payload[1]
    resp_code = telegram_payload[2]

    # "response pending" answers are always negative
    if sid != NegativeResponseId:
        return False

    # if a request SID was specified for this function, it must be
    # identical to the one which we received on the wire. (if none was
    # specified, we don't care about the received SID)
    if request_sid is not None and request_sid != rq_sid:
        return False

    # if the response code is not ResponsePending (0x78), the received
    # telegram is another error
    if resp_code != NegativeResponseCodes.ResponsePending:
        return False

    # if all of the above applies, we received a "stay tuned" response
    return True



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from .load_pdx_file import load_pdx_file
from .load_odx_d_file import load_odx_d_file

def load_file(file_name: str, enable_candela_workarounds: bool = True):
    if file_name.lower().endswith(".pdx"):
        return load_pdx_file(file_name, enable_candela_workarounds=enable_candela_workarounds)
    elif file_name.lower().endswith(".odx-d"):
        return load_odx_d_file(file_name, enable_candela_workarounds=enable_candela_workarounds)
    else:
        raise RuntimeError(f"Could not guess the file format of file '{file_name}'!")



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from enum import Enum
from typing import NamedTuple, Union


class IntervalType(Enum):
    OPEN = "OPEN"
    CLOSED = "CLOSED"
    INFINITE = "INFINITE"


class Limit(NamedTuple):
    value: Union[str, int, float, bytes]
    interval_type: IntervalType = IntervalType.CLOSED

    def complies_to_upper(self, value):
        """Checks if the value is in the range w.r.t. the upper limit.

        * If the interval type is closed, return `value <= limit.value`.
        * If the interval type is open, return `value < limit.value`.
        * If the interval type is infinite, return `True`.
        """
        if self.interval_type == IntervalType.CLOSED:
            return value <= self.value
        elif self.interval_type == IntervalType.OPEN:
            return value < self.value
        elif self.interval_type == IntervalType.INFINITE:
            return True

    def complies_to_lower(self, value):
        """Checks if the value is in the range w.r.t. the lower limit.

        * If the interval type is closed, return `limit.value <= value`.
        * If the interval type is open, return `limit.value < value`.
        * If the interval type is infinite, return `True`.
        """
        if self.interval_type == IntervalType.CLOSED:
            return self.value <= value
        elif self.interval_type == IntervalType.OPEN:
            return self.value < value
        elif self.interval_type == IntervalType.INFINITE:
            return True



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from odxtools.utils import read_description_from_odx


class CommunicationParameterRef:

    def __init__(self,
                 value,
                 id_ref,
                 description=None,
                 protocol_sn_ref=None,
                 prot_stack_sn_ref=None):
        self.value = value
        self.id_ref = id_ref
        self.description = description
        self.protocol_sn_ref = protocol_sn_ref
        self.prot_stack_sn_ref = prot_stack_sn_ref

    def __repr__(self) -> str:
        val = self.value
        if not isinstance(self.value, list):
            val = f"'{val}'"

        return f"CommunicationParameter('{self.id_ref}', value={val})"

    def __str__(self) -> str:
        val = self.value
        if not isinstance(self.value, list):
            val = f"'{val}'"

        return f"CommunicationParameter('{self.id_ref}', value={val})"

    def _python_name(self):
        return self.id_ref.replace(".", "__")


def _read_complex_value_from_odx(et_element):
    result = []
    for el in et_element.findall("*"):
        if el.tag == "SIMPLE-VALUE":
            result.append('' if el.text is None else el.text)
        else:
            result.append(_read_complex_value_from_odx(el))
    return result


def read_communication_param_ref_from_odx(et_element):
    id_ref = et_element.get("ID-REF")

    if et_element.find("SIMPLE-VALUE") is not None:
        value = et_element.find("SIMPLE-VALUE").text
    else:
        value = _read_complex_value_from_odx(et_element.find("COMPLEX-VALUE"))

    description = read_description_from_odx(et_element.find("DESC"))
    protocol_sn_ref = et_element.find(
        "PROTOCOL-SNREF").text if et_element.find("PROTOCOL-SNREF") is not None else None
    prot_stack_sn_ref = et_element.find(
        "PROT-STACK-SNREF").text if et_element.find("PROT-STACK-SNREF") is not None else None
    return CommunicationParameterRef(value,
                                     id_ref,
                                     description=description,
                                     protocol_sn_ref=protocol_sn_ref,
                                     prot_stack_sn_ref=prot_stack_sn_ref
                                     )



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import Union


class Message:
    """A CAN message with its interpretation."""

    def __init__(self,
                 coded_message: Union[bytes, bytearray],
                 service,
                 structure,
                 param_dict: dict):
        """
        Parameters
        ----------
        coded_message : bytes or bytearray
        service : DiagService
        structure : Request or Response
        param_dict : dict
        """
        self.coded_message = coded_message
        self.service = service
        self.structure = structure
        self.param_dict = param_dict

    def __getitem__(self, key: str):
        return self.param_dict[key]

    def __str__(self):
        param_string = ", ".join(map(lambda param: f"{param[0]}={repr(param[1])}",
                                     self.param_dict.items()))
        return f"{self.structure.short_name}({param_string})"

    def __repr__(self):
        return self.__str__()



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import Any, Dict, NamedTuple, Optional, Union


class EncodeState(NamedTuple):
    """Utility class to be used while encoding a message.

    While encoding parameters may update the dicts with new keys 
    but this is the only allowed change.
    In particular the coded_message is not updated in-place.
    Instead the new encode state can be constructed with::

        for p in self.parameters:
            prefix = p.encode_into_pdu(encode_state)
            encode_state = encode_state._replace(coded_message=prefix)

    """
    coded_message: bytes
    """payload that is constructed so far"""
    parameter_values: Dict[str, Any]
    """a mapping from short name to value for each parameter"""
    triggering_request: Optional[Union[bytes, bytearray]] = None
    """If encoding a response: request that triggered the response"""
    length_keys: Dict[str, int] = {}
    """Mapping from IDs to bit lengths (specified by LengthKeyParameters)"""
    is_end_of_pdu: bool = False
    """Flag whether the parameter is the last on the PDU (needed for MinMaxLengthType)"""



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from zipfile import ZipFile

from .database import Database
from .globals import logger

def load_pdx_file(pdx_file: str, enable_candela_workarounds: bool = True):
    u = ZipFile(pdx_file)
    container = Database(pdx_zip=u, enable_candela_workarounds=enable_candela_workarounds)
    logger.info(f"--- --- --- Done with parsing --- --- ---")
    return container



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import Dict, Iterable, Optional, OrderedDict, Union

from .dataobjectproperty import DataObjectProperty, DopBase
from .decodestate import DecodeState, ParameterValuePair
from .encodestate import EncodeState
from .exceptions import DecodeError, EncodeError
from .globals import logger
from .nameditemlist import NamedItemList
from .parameters import Parameter, ParameterWithDOP, read_parameter_from_odx
from .parameters import CodedConstParameter, MatchingRequestParameter, ValueParameter
from .utils import read_description_from_odx


class BasicStructure(DopBase):
    def __init__(self,
                 id,
                 short_name,
                 parameters: Iterable[Parameter],
                 long_name=None,
                 byte_size=None,
                 description=None):
        super().__init__(id, short_name, long_name=long_name, description=description)
        self.parameters = NamedItemList(lambda par: par.short_name, parameters)

        self._byte_size = byte_size

    @property
    def bit_length(self):
        if self._byte_size:
            return 8 * self._byte_size
        elif all(p.bit_length is not None for p in self.parameters):
            return sum([p.bit_length for p in self.parameters])
        else:
            return None

    def coded_const_prefix(self, request_prefix: Union[bytes, bytearray] = bytes()):
        prefix = bytearray()
        encode_state = EncodeState(
            prefix, parameter_values={}, triggering_request=request_prefix)
        for p in self.parameters:
            if isinstance(p, CodedConstParameter) and p.bit_length % 8 == 0:
                prefix = p.encode_into_pdu(encode_state)
                encode_state = EncodeState(prefix, *encode_state[1:])
            elif isinstance(p, MatchingRequestParameter):
                prefix = p.encode_into_pdu(encode_state)
                encode_state = EncodeState(prefix, *encode_state[1:])
            else:
                break
        return prefix

    def get_required_parameters(self):
        return filter(lambda p: p.is_required(), self.parameters)

    def convert_physical_to_internal(self,
                                     param_values: dict,
                                     triggering_coded_request,
                                     is_end_of_pdu=True):
        logger.debug(f"{self.short_name} encode RPC"
                     f" with params={param_values}")

        coded_rpc = bytearray()
        encode_state = EncodeState(coded_rpc,
                                   param_values,
                                   triggering_request=triggering_coded_request,
                                   is_end_of_pdu=False)

        # Keep track of the expected length. (This is just to double check and not actually needed.)
        bit_length: Optional[int] = 0
        for param in self.parameters:
            if param == self.parameters[-1]:
                # The last parameter is at the end of the PDU iff the structure itself is at the end of the PDU
                encode_state = encode_state._replace(
                    is_end_of_pdu=is_end_of_pdu)
                if isinstance(param, ValueParameter) and hasattr(param.dop, 'min_number_of_items'):
                    # The param repeats itself, making bit_length calculation invalid
                    # Temporary workaround
                    # Can not import EndOfPduField to check on its type due to circular dependency
                    bit_length = None
            coded_rpc = param.encode_into_pdu(encode_state)
            encode_state = encode_state._replace(coded_message=coded_rpc)

            # Compute expected length
            if bit_length is not None and param.bit_length is not None:
                bit_length = bit_length + param.bit_length
            else:
                bit_length = None

        # Assert that length is as expected
        if bit_length is not None:
            assert bit_length % 8 == 0, f"Length of coded structure is not divisible by 8, i.e. is not a full sequence of bytes."
            assert len(coded_rpc) == bit_length // 8, f"{self.short_name} can't encode: Actual length is {len(coded_rpc)}, computed byte length is {bit_length // 8}, computed_rpc is {coded_rpc.hex()}, \n" + '\n'.join(
                self.__message_format_lines())

        return bytearray(coded_rpc)

    def convert_physical_to_bytes(self, param_values: dict, encode_state: EncodeState, bit_position=0):
        if bit_position != 0:
            raise EncodeError("Structures must be aligned, i.e. bit_position=0, but "
                              f"{self.short_name} was passed the bit position {bit_position}")
        return self.convert_physical_to_internal(param_values,
                                                 triggering_coded_request=encode_state.triggering_request,
                                                 is_end_of_pdu=encode_state.is_end_of_pdu
                                                 )

    def convert_bytes_to_physical(self, decode_state: DecodeState, bit_position=0):
        if bit_position != 0:
            raise DecodeError("Structures must be aligned, i.e. bit_position=0, but "
                              f"{self.short_name} was passed the bit position {bit_position}")
        byte_code = decode_state.coded_message[decode_state.next_byte_position:]
        inner_decode_state = DecodeState(coded_message=byte_code,
                                         parameter_value_pairs=[],
                                         next_byte_position=0)

        for parameter in self.parameters:
            value, next_byte_position = parameter.decode_from_pdu(
                inner_decode_state)

            inner_decode_state.parameter_value_pairs.append(
                ParameterValuePair(parameter, value))
            inner_decode_state = DecodeState(coded_message=byte_code,
                                             parameter_value_pairs=inner_decode_state.parameter_value_pairs,
                                             next_byte_position=max(inner_decode_state.next_byte_position,
                                                                    next_byte_position))
        # Construct the param dict.
        # TODO: Wouldn't it be prettier if we kept the information of each parameter
        #       instead of just using the short_name as the key and "forgetting" everything else?
        param_dict = OrderedDict((pv.parameter.short_name, pv.value)
                                 for pv in inner_decode_state.parameter_value_pairs)

        return param_dict, decode_state.next_byte_position + inner_decode_state.next_byte_position

    def encode(self, coded_request=None, **params) -> bytearray:
        """
        Composes an UDS message as bytes for this service.
        Parameters:
        ----------
        coded_request: bytes
            coded request (only needed when encoding a response)
        params: dict
            Parameters of the RPC as mapping from SHORT-NAME of the parameter to the value
        """
        return self.convert_physical_to_internal(params,
                                                 triggering_coded_request=coded_request,
                                                 is_end_of_pdu=True)

    def decode(self, message: Union[bytes, bytearray]):
        # dummy decode state to be passed to convert_bytes_to_physical
        decode_state = DecodeState(coded_message=message,
                                   parameter_value_pairs=[],
                                   next_byte_position=0)
        param_values, next_byte_position = self.convert_bytes_to_physical(
            decode_state)
        if len(message) != next_byte_position:
            raise DecodeError(
                f"The message {message.hex()} is longer than could be parsed. Expected {next_byte_position} but got {len(message)}.")
        return param_values

    def parameter_dict(self):
        """
        Returns a dict with parameter short names as keys.
        The values are parameters for simple types or a nested dict for structures.
        """
        params = self.parameters
        assert all(not isinstance(p, ParameterWithDOP) or isinstance(
            p.dop, DataObjectProperty) or isinstance(p.dop, Structure) for p in self.parameters)
        param_dict: Dict[str, Union[Parameter, Dict[str, Parameter]]] = {
            p.short_name: p for p in params
            if not isinstance(p, ParameterWithDOP) or not isinstance(p.dop, Structure)
        }
        param_dict.update({
            struct_param.short_name: struct_param.dop.parameter_dict()
            for struct_param in params if isinstance(struct_param, ParameterWithDOP) and isinstance(struct_param.dop, BasicStructure)
        })
        return param_dict

    def _resolve_references(self, parent_dl, id_lookup):
        """Recursively resolve any references (odxlinks or sn-refs)
        """
        for p in self.parameters:
            if isinstance(p, ParameterWithDOP):
                p.resolve_references(parent_dl, id_lookup)

    def __message_format_lines(self, allow_unknown_lengths=False):
        # sort parameters
        sorted_params: list = list(self.parameters)  # copy list
        if all(p.byte_position is not None for p in self.parameters):
            sorted_params.sort(key=lambda p: (
                p.byte_position, 8 - p.bit_position))

        # replace structure parameters by their sub parameters
        params = []
        for p in sorted_params:
            if isinstance(p, ValueParameter) and isinstance(p.dop, BasicStructure):
                params += p.dop.parameters
            else:
                params.append(p)

        i = 0
        byte = 0
        # needs to be one larger than the maximum digit length of a byte number
        indent_for_byte_numbering = 5 * " "
        formatted_lines = [indent_for_byte_numbering +
                           "".join(f"   {7-bit}  " for bit in range(8))]

        breakpoint = 0  # absolute bit position where the next parameter starts

        divide_string = indent_for_byte_numbering + 8 * "+-----" + "+"

        error = False
        next_line = ""
        while i <= len(params) and not error:  # For each byte
            if 8 * byte == breakpoint and i == len(params):
                # If we have formatted the last parameter, we're done.
                break

            formatted_lines.append(f"{divide_string}")
            if breakpoint // 8 - byte > 5:
                curr_param = params[i-1].short_name
                formatted_lines.append(
                    indent_for_byte_numbering + f"  ... {breakpoint // 8 - byte} bytes belonging to {curr_param} ... ")
                byte += breakpoint // 8 - byte
                continue

            next_line = f"{(len(indent_for_byte_numbering) - 1 - len(str(byte))) * ' '}{byte} "

            for bit in range(8):
                assert 8 * byte + bit <= breakpoint

                if 8 * byte + bit == breakpoint:
                    # END-OF-PDU fields do not exhibit a fixed bit
                    # length, so they need special treatment here
                    dct = None
                    if hasattr(params[i], 'dop'):
                        dop = params[i].dop
                        if hasattr(dop, 'diag_coded_type'):
                            dct = dop.diag_coded_type

                    if dct is not None and dct.dct_type == 'MIN-MAX-LENGTH-TYPE':
                        name = params[i].short_name + " ("
                        if dct.termination == "END-OF-PDU":
                            name += "End of PDU, "
                        name += f"{dct.min_length}..{dct.max_length} bytes"
                        name += ")"
                        next_line += "| " + name

                        i += 1

                        # adding 8 is is a bit hacky here, but hey, it
                        # works ...
                        breakpoint += 8

                        break

                    elif not params[i].bit_length and not allow_unknown_lengths:
                        # The bit length is not set for the current
                        # parameter, i.e. it was either not specified
                        # or the parameter is of variable length and
                        # has an type which is not handled above. In
                        # this case, stop trying.
                        error = True
                        break
                    else:
                        breakpoint += params[i].bit_length or (
                            allow_unknown_lengths and 8)
                        name = params[i].short_name + \
                            f" ({params[i].bit_length or 'Unknown'} bits)"
                        next_line += "| " + name

                    i += 1

                    if byte == breakpoint // 8:
                        char_pos = bit * 6 + 2 + len(name)
                        width_of_line = (breakpoint % 8) * 6
                        if char_pos < width_of_line:
                            next_line += " " * \
                                (width_of_line - char_pos) + "|"
                        # start next line (belongs to same byte)
                        formatted_lines.append(next_line)
                        # fill next line with white spaces upto the bit where next parameter starts
                        next_line = indent_for_byte_numbering + \
                            (bit + 1) * 6 * " "
                    else:
                        char_pos = 2 + bit * 6 + len(name)
                        width_of_line = 8 * 6
                        if char_pos < width_of_line:
                            next_line += " " * \
                                (width_of_line - char_pos) + "|"
                        break
                else:
                    if bit == 0:
                        next_line += "|" + 5 * " "
                    elif bit == 7:
                        next_line += 6 * " " + "|"
                    else:
                        next_line += 6 * " "

            formatted_lines.append(next_line)
            next_line = ""

            byte += 1

        if not error:
            formatted_lines.append(divide_string)
            return formatted_lines
        else:
            return None

    def print_message_format(self, indent: int = 5, allow_unknown_lengths=False):
        """
        Print a description of the message format to `stdout`.
        """

        message_as_lines = self.__message_format_lines(
            allow_unknown_lengths=allow_unknown_lengths)
        if message_as_lines is not None:
            print(f"{indent * ' '}" +
                  f"\n{indent * ' '}".join(message_as_lines))
        else:
            print("Sorry, couldn't pretty print message layout. :(")
        for p in self.parameters:
            print(indent * ' ' +
                  str(p).replace("\n", f"\n{indent * ' '}"))


class Structure(BasicStructure):
    def __init__(self, id, short_name, parameters, long_name=None, byte_size=None, description=None):
        super().__init__(id, short_name, parameters,
                         long_name=long_name, description=description)

        self.parameters = parameters
        self.basic_structure = BasicStructure(id, short_name, parameters,
                                              long_name=long_name, byte_size=byte_size, description=description)

    def __repr__(self) -> str:
        return f"Structure('{self.short_name}', byte_size={self._byte_size})"

    def __str__(self) -> str:
        params = "[\n" + "\n".join([" " + str(p).replace("\n", "\n ")
                                   for p in self.parameters]) + "\n]"

        return \
            f"Structure '{self.short_name}': " + \
            f"Byte size={self._byte_size}, " + \
            f"Parameters={params}"


class Request(BasicStructure):
    def __init__(self, id, short_name, parameters, long_name=None, description=None):
        super().__init__(id, short_name, parameters,
                         long_name=long_name, description=description)

    def __repr__(self) -> str:
        return f"Request('{self.short_name}')"

    def __str__(self) -> str:
        return f"Request('{self.short_name}')"


class Response(BasicStructure):
    def __init__(self, id, short_name, parameters, long_name=None, response_type=None, description=None):
        super().__init__(id, short_name, parameters,
                         long_name=long_name, description=description)
        self.response_type = "POS-RESPONSE" if response_type == "POS-RESPONSE" else "NEG-RESPONSE"

    def encode(self, coded_request: bytearray = None, **params):
        logger.info(f"Compose response message to the request {coded_request}")
        if coded_request is not None:
            # Extract MATCHING-REQUEST-PARAMs from the coded request
            for param in self.parameters:
                if param.parameter_type == "MATCHING-REQUEST-PARAM":
                    logger.info(
                        f"set matching request param value {param.short_name}")
                    byte_pos = param.request_byte_position
                    byte_length = param.byte_length

                    val = coded_request[byte_pos:byte_pos+byte_length]
                    params[param.short_name] = val

        return super().encode(coded_request=coded_request, **params)

    def __repr__(self) -> str:
        return f"Response('{self.short_name}')"

    def __str__(self) -> str:
        return f"Response('{self.short_name}')"


def read_structure_from_odx(et_element) -> Union[Structure, Request, Response, None]:
    id = et_element.get("ID")
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.find("LONG-NAME").text
    description = read_description_from_odx(et_element.find("DESC"))
    parameters = [read_parameter_from_odx(et_parameter)
                  for et_parameter in et_element.iterfind("PARAMS/PARAM")]

    res: Union[Structure, Request, Response, None]
    if et_element.tag == "REQUEST":
        res = Request(
            id,
            short_name,
            parameters=parameters,
            long_name=long_name,
            description=description
        )
    elif et_element.tag in ["POS-RESPONSE", "NEG-RESPONSE"]:
        res = Response(
            id,
            short_name,
            response_type=et_element.tag,
            parameters=parameters,
            long_name=long_name,
            description=description
        )
    elif et_element.tag == "STRUCTURE":
        byte_size = int(et_element.find(
            "BYTE-SIZE").text) if et_element.find("BYTE-SIZE") is not None else None
        res = Structure(
            id,
            short_name,
            parameters=parameters,
            byte_size=byte_size,
            long_name=long_name,
            description=description
        )
    else:
        res = None
        logger.critical(
            f"Did not recognize structure {et_element.tag} {short_name}")
    return res



# SPDX-License-Identifier: MIT
# Copyright (c) 2021-2022 MBition GmbH

import argparse

from . import list
from . import browse
from . import snoop
from . import find

from ..version import __version__ as odxtools_version

def start_cli():
    argparser = argparse.ArgumentParser(
        description="\n".join([
            "Utilities to interact with automotive diagnostic descriptions based on the ODX standard.",
            "",
            "Examples:",
            "  For printing all services use:",
            "   odxtools list ./path/to/database.pdx --services",
            "  For browsing the data base and encoding messages use:",
            "   odxtools browse ./path/to/database.pdx"
        ]),
        prog="odxtools",
        formatter_class=argparse.RawTextHelpFormatter)

    argparser.add_argument("--version", required=False,
                           action='store_true',
                           help="Print the odxtools version")

    subparsers = argparser.add_subparsers(
        help='Select a subcommand',
        dest="subparser_name"
    )

    list.add_subparser(subparsers)
    browse.add_subparser(subparsers)
    snoop.add_subparser(subparsers)
    find.add_subparser(subparsers)

    args = argparser.parse_args()  # deals with the help message handling

    if args.version:
        print(f"odxtools {odxtools_version}")
        exit()
    if args.subparser_name is None:
        argparser.print_usage()
        exit()

    if args.subparser_name == "list":
        list.run(args)
    elif args.subparser_name == "browse":
        browse.run(args)
    elif args.subparser_name == "snoop":
        snoop.run(args)
    elif args.subparser_name == "find":
        find.run(args)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import Any, Dict, List, NamedTuple, Union


class ParameterValuePair(NamedTuple):
    parameter: Any  # This should be a Parameter but that would lead to a cyclic import
    value: Union[str, int, bytes, bytearray, Dict]


class DecodeState(NamedTuple):
    """Utility class to be used while decoding a message."""
    coded_message: Union[bytes, bytearray]
    """bytes to be decoded"""
    parameter_value_pairs: List[ParameterValuePair]
    """values of already decoded parameters"""
    next_byte_position: int
    """Position of the next parameter if its position is not specified in ODX"""



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from .parameterbase import Parameter
from .parameterwithdop import ParameterWithDOP

from .codedconstparameter import CodedConstParameter
from .dynamicparameter import DynamicParameter
from .lengthkeyparameter import LengthKeyParameter
from .matchingrequestparameter import MatchingRequestParameter
from .nrcconstparameter import NrcConstParameter
from .physicalconstantparameter import PhysicalConstantParameter
from .reservedparameter import ReservedParameter
from .systemparameter import SystemParameter
from .tableentryparameter import TableEntryParameter
from .tablekeyparameter import TableKeyParameter
from .tablestructparameter import TableStructParameter
from .valueparameter import ValueParameter

from .readparameter import read_parameter_from_odx



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..exceptions import DecodeError

from .parameterwithdop import ParameterWithDOP


class PhysicalConstantParameter(ParameterWithDOP):
    def __init__(self,
                 short_name,
                 physical_constant_value,
                 **kwargs):
        super().__init__(short_name, parameter_type="PHYS-CONST",
                         **kwargs)
        assert physical_constant_value is not None
        self._physical_constant_value = physical_constant_value

    @property
    def physical_constant_value(self):
        # Cast to physical type
        return self.dop.physical_type.base_data_type.from_string(self._physical_constant_value)

    def is_required(self):
        return False

    def is_optional(self):
        return False

    def get_coded_value(self):
        return self.dop.convert_physical_to_internal(self.physical_constant_value)

    def get_coded_value_as_bytes(self, encode_state: EncodeState):
        assert self.dop is not None, "Reference to DOP is not resolved"
        if self.short_name in encode_state.parameter_values \
                and encode_state.parameter_values[self.short_name] != self.physical_constant_value:
            raise TypeError(f"The parameter '{self.short_name}' is constant {self.physical_constant_value}"
                            " and thus can not be changed.")
        return self.dop.convert_physical_to_bytes(self.physical_constant_value, encode_state, bit_position=self.bit_position)

    def decode_from_pdu(self, decode_state: DecodeState):
        # Decode value
        phys_val, next_byte_position = super().decode_from_pdu(decode_state)

        # Check if decoded value matches expected value
        if phys_val != self.physical_constant_value:
            raise DecodeError(
                f"Physical constant parameter does not match! "
                f"The parameter {self.short_name} expected physical value {self.physical_constant_value!r} but got {phys_val!r} "
                f"at byte position {next_byte_position} "
                f"in coded message {decode_state.coded_message.hex()}."
            )
        return phys_val, next_byte_position

    def __repr__(self) -> str:
        repr_str = f"PhysicalConstantParameter(short_name='{self.short_name}', physical_constant_value={self.physical_constant_value}"
        if self.long_name is not None:
            repr_str += f", long_name='{self.long_name}'"
        if self.byte_position is not None:
            repr_str += f", byte_position='{self.byte_position}'"
        if self.bit_position:
            repr_str += f", bit_position='{self.bit_position}'"
        if self.semantic is not None:
            repr_str += f", semantic='{self.semantic}'"
        if self.dop_ref is not None:
            repr_str += f", dop_ref={repr(self.dop_ref)}"
        elif self.dop_snref is not None:
            repr_str += f", dop_snref={repr(self.dop_snref)}"
        if self.description is not None:
            repr_str += f", description='{' '.join(self.description.split())}'"
        return repr_str + ")"



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from .parameterbase import Parameter


class TableEntryParameter(Parameter):
    def __init__(self,
                 short_name,
                 target,
                 table_row_ref,
                 long_name=None,
                 byte_position=None,
                 bit_position=0,
                 semantic=None,
                 description=None):
        super().__init__(
            short_name=short_name,
            long_name=long_name,
            byte_position=byte_position,
            bit_position=bit_position,
            parameter_type="TABLE-ENTRY",
            semantic=semantic,
            description=description
        )
        assert target in ["KEY", "STRUCT"]
        self.target = target
        self.table_row_ref = table_row_ref

    def is_required(self):
        raise NotImplementedError(
            "TableKeyParameter.is_required is not implemented yet.")

    def is_optional(self):
        raise NotImplementedError(
            "TableKeyParameter.is_optional is not implemented yet.")

    def get_coded_value(self):
        raise NotImplementedError(
            "Encoding a TableKeyParameter is not implemented yet.")

    def get_coded_value_as_bytes(self):
        raise NotImplementedError(
            "Encoding a TableKeyParameter is not implemented yet.")

    def decode_from_pdu(self, coded_message, default_byte_position=None):
        raise NotImplementedError(
            "Decoding a TableKeyParameter is not implemented yet.")



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from .parameterwithdop import ParameterWithDOP


class SystemParameter(ParameterWithDOP):
    def __init__(self,
                 short_name,
                 sysparam,
                 dop_ref=None,
                 dop_snref=None,
                 long_name=None,
                 byte_position=None,
                 bit_position=0,
                 semantic=None,
                 description=None):
        super().__init__(short_name,
                         parameter_type="SYSTEM",
                         dop_ref=dop_ref,
                         dop_snref=dop_snref,
                         long_name=long_name,
                         byte_position=byte_position,
                         bit_position=bit_position,
                         semantic=semantic,
                         description=description)
        self.sysparam = sysparam

    def is_required(self):
        raise NotImplementedError(
            "SystemParameter.is_required is not implemented yet.")

    def is_optional(self):
        raise NotImplementedError(
            "SystemParameter.is_optional is not implemented yet.")

    def get_coded_value(self):
        raise NotImplementedError(
            "Encoding a SystemParameter is not implemented yet.")

    def get_coded_value_as_bytes(self):
        raise NotImplementedError(
            "Encoding a SystemParameter is not implemented yet.")

    def decode_from_pdu(self, coded_message, default_byte_position=None):
        raise NotImplementedError(
            "Decoding a SystemParameter is not implemented yet.")

    def __repr__(self):
        repr_str = f"SystemParameter(short_name='{self.short_name}', sysparam='{self.sysparam}'"
        if self.long_name is not None:
            repr_str += f", long_name='{self.long_name}'"
        if self.byte_position is not None:
            repr_str += f", byte_position='{self.byte_position}'"
        if self.bit_position:
            repr_str += f", bit_position='{self.bit_position}'"
        if self.semantic is not None:
            repr_str += f", semantic='{self.semantic}'"
        if self.dop_ref is not None:
            repr_str += f", dop_ref='{self.dop_ref}'"
        if self.dop_snref is not None:
            repr_str += f", dop_snref='{self.dop_snref}'"
        if self.description is not None:
            repr_str += f", description='{' '.join(self.description.split())}'"
        return repr_str + ")"



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from ..globals import logger

from .compumethodbase import CompuMethod


class TabIntpCompuMethod(CompuMethod):
    def __init__(self,
                 internal_type,
                 physical_type):
        super().__init__(internal_type, physical_type, "TAB-INTP")
        logger.debug("Created table interpolation compu method!")
        logger.warning(
            "TODO: Implement table interpolation compu method properly!")

    def convert_physical_to_internal(self, physical_value):
        return self.internal_type.make_from(physical_value)

    def convert_internal_to_physical(self, internal_value):
        return self.physical_type.make_from(internal_value)

    def is_valid_physical_value(self, physical_value):
        return True

    def is_valid_internal_value(self, internal_value):
        return True



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import Any, Dict
from xml.etree import ElementTree
from itertools import chain
from zipfile import ZipFile

from .diaglayer import DiagLayer, read_diag_layer_container_from_odx
from .globals import logger
from .nameditemlist import NamedItemList

class Database:
    """This class internalizes the diagnostic database for various ECUs
    described by a collection of ODX files which are usually collated
    into a single PDX file.
    """

    def __init__(self,
                 pdx_zip: ZipFile = None,
                 odx_d_file_name: str = None,
                 enable_candela_workarounds: bool = True):

        self._id_lookup: Dict[str, Any] = {}
        dlc_elements = []

        if pdx_zip is None and odx_d_file_name is None:
            # create an empty database object
            return

        if pdx_zip is not None and odx_d_file_name is not None:
            raise TypeError("The 'pdx_zip' and 'odx_d_file_name' parameters are mutually exclusive")

        if pdx_zip is not None:
            names = list(pdx_zip.namelist())
            names.sort()
            for zip_member in names:
                if zip_member.endswith(".odx-d"):
                    logger.info(f"Processing the file {zip_member}")
                    d = pdx_zip.read(zip_member)
                    root = ElementTree.fromstring(d)

                    dlc_elements.append(root.find("DIAG-LAYER-CONTAINER"))

            tmp = [
                read_diag_layer_container_from_odx(dlc_el, enable_candela_workarounds=enable_candela_workarounds) \
                  for dlc_el in dlc_elements
            ]
            self._diag_layer_containers = NamedItemList(lambda x: x.short_name, tmp)
            self._diag_layer_containers.sort(key=lambda x: x.short_name)
            self.finalize_init()

        elif odx_d_file_name is not None:
            dlc_element = ElementTree.parse(odx_d_file_name).find("DIAG-LAYER-CONTAINER")

            self._diag_layer_containers = \
                NamedItemList(lambda x: x.short_name,
                              [read_diag_layer_container_from_odx(dlc_element)])
            self.finalize_init()

    def finalize_init(self):
        # Create wrapper objects
        self._diag_layers = NamedItemList(
            lambda dl: dl.short_name,
            chain(*(dlc.diag_layers for dlc in self.diag_layer_containers)))
        self._ecus = NamedItemList(
            lambda ecu: ecu.short_name,
            chain(*(dlc.ecu_variants for dlc in self.diag_layer_containers)))

        # Build id_lookup
        self._id_lookup = {}

        for dlc in self.diag_layer_containers:
            self.id_lookup.update(dlc._build_id_lookup())

        for dl in self.diag_layers:
            self.id_lookup.update(dl._build_id_lookup())

        # Resolve references
        for dlc in self.diag_layer_containers:
            dlc._resolve_references(self.id_lookup)

        for dl_type_name in ["ECU-SHARED-DATA", "PROTOCOL", "FUNCTIONAL-GROUP", "BASE-VARIANT", "ECU-VARIANT"]:
            for dl in self.diag_layers:
                if dl.variant_type == dl_type_name:
                    dl._resolve_references(self.id_lookup)

    @property
    def id_lookup(self) -> dict:
        """A map from id to object"""
        return self._id_lookup

    @property
    def ecus(self) -> NamedItemList[DiagLayer]:
        """ECU-variants defined in the data base"""
        return self._ecus

    @property
    def diag_layers(self) -> NamedItemList[DiagLayer]:
        """all diagnostic layers defined in the data base"""
        return self._diag_layers

    @property
    def diag_layer_containers(self):
        return self._diag_layer_containers

    @diag_layer_containers.setter
    def diag_layer_containers(self, value):
        self._diag_layer_containers = value



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


import abc
from typing import Optional

from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..globals import logger


class Parameter(abc.ABC):
    def __init__(self,
                 short_name,
                 parameter_type,
                 long_name=None,
                 byte_position=None,
                 bit_position=0,
                 semantic=None,
                 description=None):
        self.short_name: str = short_name
        self.long_name: Optional[str] = long_name
        self.byte_position: Optional[int] = byte_position
        self.bit_position: int = bit_position
        self.parameter_type: str = parameter_type
        self.semantic: Optional[str] = semantic
        self.description: Optional[str] = description

    @property
    def bit_length(self) -> Optional[int]:
        return None

    @abc.abstractclassmethod
    def is_required(self):
        pass

    @abc.abstractclassmethod
    def is_optional(self):
        pass

    @abc.abstractclassmethod
    def get_coded_value(self):
        pass

    @abc.abstractclassmethod
    def get_coded_value_as_bytes(self, encode_state: EncodeState) -> bytes:
        """Get the coded value of the parameter given the encode state.
        Note that this method is called by `encode_into_pdu`.
        """
        pass

    @abc.abstractclassmethod
    def decode_from_pdu(self, decode_state: DecodeState):
        """Decode the parameter value from the coded message.

        If the paramter does have a byte position property, the coded bytes the parameter covers are extracted
        at this byte position and the function parameter `default_byte_position` is ignored.

        If the parameter does not have a byte position and a byte position is passed,
        the bytes are extracted at the byte position given by the argument `default_byte_position`.

        If the parameter does not have a byte position and the argument `default_byte_position` is None,
        this function throws a `DecodeError`.

        Parameters
        ----------
        decode_state : DecodeState
            The decoding state containing
            * the byte message to be decoded
            * the parameter values that are already decoded
            * the next byte position that is used iff the parameter does not specify a byte position

        Returns
        -------
        ParameterValuePair | List[ParameterValuePair]
            the decoded parameter value (the type is defined by the DOP)
        int
            the next byte position after the extracted parameter
        """
        pass

    def encode_into_pdu(self, encode_state: EncodeState) -> bytearray:
        """Insert the encoded value of a parameter into the coded RPC.

        If the byte position of the parameter is not defined,
        the byte code is appended to the coded RPC.

        Technical note for subclasses:
        The default implementation tries to compute the coded value
        via `self.get_coded_value_as_bytes(encoded_state)` and inserts it into
        the PDU. Thus it suffices to overwrite `get_coded_value_as_bytes()`.

        Parameters:
        ----------
        encode_state: EncodeState, i.e. a named tuple with attributes
            * coded_message: bytes, the message encoded so far
            * parameter_values: List[ParameterValuePairs]
            * triggering_coded_request: bytes

        Returns:
        -------
        bytearray
            the updated, coded message after encoding the parameter into it
        """
        byte_value = self.get_coded_value_as_bytes(encode_state)

        old_rpc = encode_state.coded_message
        new_rpc = bytearray(old_rpc)

        if self.byte_position is not None:
            byte_position = self.byte_position
        else:
            byte_position = len(old_rpc)

        min_length = byte_position + len(byte_value)
        if len(old_rpc) < min_length:
            # Make byte code longer if necessary
            new_rpc += bytearray([0] * (min_length - len(old_rpc)))
        for byte_idx_val, byte_idx_rpc in enumerate(range(byte_position, byte_position + len(byte_value))):
            # insert byte value
            assert new_rpc[byte_idx_rpc] & byte_value[byte_idx_val] == 0, "Bytes are already set!"
            new_rpc[byte_idx_rpc] |= byte_value[byte_idx_val]

        logger.debug(f"Param {self.short_name} inserts"
                     f" 0x{byte_value.hex()} at byte pos {byte_position}")
        return new_rpc

    def _as_dict(self):
        """
        Mostly for pretty printing purposes (specifically not for reconstructing the object)
        """
        d = {
            'short_name': self.short_name,
            'type': self.parameter_type,
            'semantic': self.semantic
        }
        if self.byte_position is not None:
            d['byte_position'] = self.byte_position
        if self.bit_position:
            d['bit_position'] = self.bit_position
        return d

    def __repr__(self):
        repr_str = f"Parameter(parameter_type='{self.parameter_type}', short_name='{self.short_name}'"
        if self.long_name is not None:
            repr_str += f", long_name='{self.long_name}'"
        if self.byte_position is not None:
            repr_str += f", byte_position='{self.byte_position}'"
        if self.bit_position:
            repr_str += f", bit_position='{self.bit_position}'"
        if self.semantic is not None:
            repr_str += f", semantic='{self.semantic}'"
        if self.description is not None:
            repr_str += f", description='{' '.join(self.description.split())}'"
        return repr_str + ")"

    def __str__(self):
        # create list of all parameters. 'short_name' ought to be
        # first, so it needs special treatment...
        param_descs = [f"short_name='{self.short_name}'"]
        for (key, val) in self._as_dict().items():
            if key == "short_name":
                continue
            elif isinstance(val, str):
                param_descs.append(f"{key}='{val}'")
            else:
                param_descs.append(f"{key}={val}")

        return f"Parameter({', '.join(param_descs)})"



#! /usr/bin/env python3
#
# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from setuptools import setup
from setuptools import find_packages
import pathlib
import re

cwd = pathlib.Path(__file__).parent

readme_text = (cwd / "README.md").read_text()
requires_list = [
    "bitstruct >= 6.0.0",
    "argparse_addons",
    "PyInquirer",
    "jinja2",
    "python-can < 4.0",
    "can-isotp",
    "markdownify",
]
version_string = re.search(r"^__version__ = '(.*)'$",
                           (cwd / "odxtools" / "version.py").read_text(),
                           re.MULTILINE).group(1)

setup(name='odxtools',
      version=version_string,
      description='Utilities to work with the automotive diagnostics standard ODX.',
      long_description=readme_text,
      long_description_content_type='text/markdown',
      author='Katrin Bauer, Andreas Lauser',
      author_email='katrin.b.bauer@mbition.io, andreas.lauser@mbition.io',
      url='https://github.com/Daimler/odxtools',
      license='MIT',
      project_urls={
          'Bug Tracker': 'https://github.com/Daimler/odxtools/issues',
      },
      classifiers=[
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.8',
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: POSIX'
      ],
      keywords=['can', 'can bus', 'odx', 'pdx', 'obd', 'uds', 'automotive', 'diagnostics'],
      packages=find_packages(exclude=['tests']),
      python_requires='>=3.8',
      include_package_data=True,
      install_requires=requires_list,
      test_suite='tests',
      entry_points={
          'console_scripts': ['odxtools=odxtools.__init__:_main']
      })



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from typing import List

from ..exceptions import DecodeError
from ..odxtypes import DataType

from .compumethodbase import CompuMethod
from .compuscale import CompuScale


class TexttableCompuMethod(CompuMethod):
    def __init__(self, internal_to_phys: List[CompuScale], internal_type):
        super().__init__(internal_type, DataType.A_UNICODE2STRING, "TEXTTABLE")
        self.internal_to_phys = internal_to_phys

        assert all(scale.lower_limit is not None or scale.upper_limit is not None
                   for scale in self.internal_to_phys), "Text table compu method doesn't have expected format!"

    def convert_physical_to_internal(self, physical_value):
        scale = next(filter(
            lambda scale: scale.compu_const == physical_value, self.internal_to_phys), None)
        if scale is not None:
            res = scale.compu_inverse_value if scale.compu_inverse_value is not None else scale.lower_limit.value
            assert self.internal_type.isinstance(res)
            return res

    def __is_internal_in_scale(self, internal_value, scale: CompuScale):
        assert scale.lower_limit is not None
        if scale.lower_limit.value is None:
            return internal_value == scale.lower_limit.value
        else:
            return scale.lower_limit.value <= internal_value and internal_value <= scale.lower_limit.value

    def convert_internal_to_physical(self, internal_value):
        try:
            scale = next(filter(lambda scale: self.__is_internal_in_scale(internal_value, scale),
                                self.internal_to_phys))
        except StopIteration:
            raise DecodeError(
                f"Texttable compu method could not decode {internal_value} to string.")
        return scale.compu_const

    def is_valid_physical_value(self, physical_value):
        return physical_value in map(lambda x: x.compu_const, self.internal_to_phys)

    def is_valid_internal_value(self, internal_value):
        return any(self.__is_internal_in_scale(internal_value, scale) for scale in self.internal_to_phys)

    def get_valid_physical_values(self):
        return list(map(lambda x: x.compu_const, self.internal_to_phys))



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import argparse
from typing import Dict, List

from ._print_utils import print_diagnostic_service
from ..database import Database
from ..service import DiagService
from . import _parser_utils


def get_display_value(v, _param):
    import binascii

    if isinstance(v, bytes):
        return binascii.hexlify(v, ' ', 1).decode('utf-8')
    elif isinstance(v, int):
        return f"{v} ({hex(v)})"
    else:
        return v


def print_decoded_message(service: DiagService, message: bytes):
    decoded = service.decode_message(message)
    print(f"\nDecoded {decoded.structure}:")
    for k, v in decoded.param_dict.items():
        param = decoded.structure.parameter_dict()[k]
        print(f"\t{k}: {get_display_value(v, param)}")
    pass


def print_summary(odxdb: Database,
                  ecu_variants=None,
                  data=None,
                  service_names=None,
                  decode=False,
                  print_params=False,
                  allow_unknown_bit_lengths=False):
    ecu_names = ecu_variants if ecu_variants else [
        ecu.short_name for ecu in odxdb.ecus
    ]
    services: Dict[DiagService, List[str]] = {}
    for ecu_name in ecu_names:
        ecu = odxdb.ecus[ecu_name]
        if not ecu:
            print(f"The ecu variant '{ecu_name}' could not be found!")
            continue
        if data:
            found_services = ecu._find_services_for_uds(data)
            for found_service in found_services:
                ecu_names = services.get(found_service, [])
                ecu_names.append(ecu_name)
                services[found_service] = ecu_names

        if service_names:
            for service_name_search in service_names:
                for service in ecu.services:
                    if service_name_search.lower() in service.short_name.lower():
                        ecu_names = services.get(service, [])
                        ecu_names.append(ecu_name)
                        services[service] = ecu_names

    for service, ecu_names in services.items():
        display_names = ', '.join(ecu_names)
        filler = str.ljust('', len(display_names), '=')
        print(f"\n{filler}")
        print(f"{', '.join(ecu_names)}")
        print(f"{filler}\n\n")
        print_diagnostic_service(service,
                                 print_params=print_params,
                                 allow_unknown_bit_lengths=allow_unknown_bit_lengths,
                                 print_pre_condition_states=True,
                                 print_state_transitions=True,
                                 print_audiences=True)
        if decode:
            print_decoded_message(service, data)


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        "find",
        description="\n".join([
            "Find & print services by hex-data, or name, can also decodes requests",
            "",
            "Examples:",
            "  For displaying the service associated with the request 10 01:",
            "    odxtools find ./path/to/database.pdx -d 10 01",
            "  For displaying the service associated with the request 10 01, and decoding it:",
            "    odxtools find ./path/to/database.pdx -D 10 01",
            "  For displaying the services associated with the partial name 'Reset' without details:",
            "    odxtools find ./path/to/database.pdx -s \"Reset\" --no-details",
            "  For more information use:",
            "    odxtools find -h"
        ]),
        help="Find & print services by hex-data, or name. Can also decode the request.",
        formatter_class=argparse.RawTextHelpFormatter)
    _parser_utils.add_pdx_argument(parser)

    parser.add_argument("-v", "--variants", nargs='+', metavar="VARIANT",
                        required=False, help="Specifies which ecu variants should be included.", default="all")

    parser.add_argument("-d", "--data", nargs='*', default=None, metavar="DATA",
                        required=False, help="Print a list of diagnostic services associated with the hex request.")

    parser.add_argument("-D", "--decode", nargs='*', default=None, metavar="DECODE",
                        required=False, help="Print a list of diagnostic services associated with the hex request and decode the request.")

    parser.add_argument("-s", "--service-names", nargs='*', default=None, metavar="SERVICES",
                        required=False, help="Print a list of diagnostic services partially matching given service names")

    parser.add_argument("-nd", "--no-details", action='store_false', required=False, help="Don't show all service details")

    parser.add_argument("-ro", "--relaxed-output", action='store_true', required=False, help="Relax output formatting rules (allow unknown bitlengths for ascii representation)")


def hex_to_binary(data):
    import binascii
    return binascii.unhexlify(''.join(data).replace(' ', ''))


def run(args):
    odxdb = _parser_utils.load_file(args)

    variants = args.variants if args.variants else None

    data = hex_to_binary(args.data) if args.data else hex_to_binary(args.decode) if args.decode else None
    decode = True if args.decode else False

    print_summary(odxdb,
                  ecu_variants=None if variants == "all" else variants,
                  data=data,
                  decode=decode,
                  service_names=args.service_names,
                  print_params=args.no_details,
                  allow_unknown_bit_lengths=args.relaxed_output)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import abc
from typing import Union

from .odxtypes import DataType
from .exceptions import DecodeError, EncodeError
from .globals import xsi, logger
from .decodestate import DecodeState
from .encodestate import EncodeState

import bitstruct

ODX_TYPE_TO_FORMAT_LETTER = {
    DataType.A_INT32: 's',
    DataType.A_UINT32: 'u',
    DataType.A_FLOAT32: 'f',
    DataType.A_FLOAT64: 'f',
    DataType.A_BYTEFIELD: 'r',
    DataType.A_UNICODE2STRING: 'r',  # UTF-16 strings must be converted explicitly
    DataType.A_ASCIISTRING: 't',
    DataType.A_UTF8STRING: 't'
}


class DiagCodedType(abc.ABC):
    def __init__(self,
                 base_data_type: Union[str, DataType],
                 dct_type: str,
                 base_type_encoding=None,
                 is_highlow_byte_order: bool = True):
        self.base_data_type = DataType(base_data_type)
        self.dct_type = dct_type
        self.base_type_encoding = base_type_encoding
        self.is_highlow_byte_order = is_highlow_byte_order

    def _extract_internal(self,
                          coded_message,
                          byte_position,
                          bit_position,
                          bit_length,
                          base_data_type: DataType,
                          is_highlow_byte_order,
                          bit_mask=None):
        """Extract the internal value.

        Helper method for `DiagCodedType.convert_bytes_to_internal`.
        """
        # If the bit length is zero, return "empty" values of each type
        if bit_length == 0:
            return base_data_type.as_python_type()(), byte_position

        byte_length = (bit_length + bit_position + 7) // 8
        if byte_position + byte_length > len(coded_message):
            raise DecodeError(
                f"Expected a longer message."
            )
        next_byte_position = byte_position + byte_length
        extracted_bytes = coded_message[byte_position:next_byte_position]

        # TODO: Apply bit mask, etc.
        if bit_mask is not None:
            raise NotImplementedError(
                f"Don't know how to handle bit_mask={bit_mask}.")

        # Apply byteorder
        if not is_highlow_byte_order and base_data_type not in [DataType.A_UNICODE2STRING, DataType.A_BYTEFIELD, DataType.A_ASCIISTRING, DataType.A_UTF8STRING]:
            extracted_bytes = extracted_bytes[::-1]

        format_letter = ODX_TYPE_TO_FORMAT_LETTER[base_data_type]
        padding = 8 * byte_length - (bit_length + bit_position)
        internal_value = bitstruct.unpack_from(
            f"{format_letter}{bit_length}", extracted_bytes, offset=padding)[0]

        if base_data_type == DataType.A_UNICODE2STRING:
            # Convert bytes to string with utf-16 decoding
            if is_highlow_byte_order:
                internal_value = internal_value.decode("utf-16-be")
            else:
                internal_value = internal_value.decode("utf-16-le")

        return internal_value, next_byte_position

    def _to_bytes(self, internal_value, bit_position, bit_length, base_data_type, is_highlow_byte_order, bit_mask=None):
        """Convert the internal_value to bytes."""
        # Check that bytes and strings actually fit into the bit length
        if base_data_type in [DataType.A_BYTEFIELD] and 8 * len(internal_value) > bit_length:
            raise EncodeError(f"The bytefield {internal_value.hex()} is too large."
                              f" The maximum byte length is {bit_length//8}.")
        if base_data_type in [DataType.A_ASCIISTRING, DataType.A_UTF8STRING] and 8 * len(internal_value) > bit_length:
            raise EncodeError(f"The string {repr(internal_value)} is too large."
                              f" The maximum number of characters is {bit_length//8}.")
        if base_data_type in [DataType.A_UNICODE2STRING] and 16 * len(internal_value) > bit_length:
            raise EncodeError(f"The string {repr(internal_value)} is too large."
                              f" The maximum number of characters is {bit_length//16}.")

        # If the bit length is zero, return empty bytes
        if bit_length == 0:
            if base_data_type in [DataType.A_INT32, DataType.A_UINT32, DataType.A_FLOAT32, DataType.A_FLOAT64] and base_data_type != 0:
                raise EncodeError(
                    f"The number {repr(internal_value)} cannot be encoded into {bit_length} bits.")
            return bytes()

        char = ODX_TYPE_TO_FORMAT_LETTER[base_data_type]

        # The coded byte is divided into (0..0)(value)(0..0) with bit lengths (left_pad)(bit_length)(bit_position)
        offset = (8 - ((bit_length + bit_position) % 8)) % 8
        assert 0 <= offset and offset < 8 and (
            offset + bit_length + bit_position) % 8 == 0, f"Computational mistake, offset={offset}"
        left_pad = f'p{offset}' if offset > 0 else ''

        # Convert string to bytes with utf-16 encoding
        if base_data_type == DataType.A_UNICODE2STRING:
            if is_highlow_byte_order:
                internal_value = internal_value.encode("utf-16-be")
            else:
                internal_value = internal_value.encode("utf-16-le")

        code = bitstruct.pack(f'{left_pad}{char}{bit_length}', internal_value)

        if not is_highlow_byte_order and base_data_type not in [DataType.A_UNICODE2STRING, DataType.A_BYTEFIELD, DataType.A_ASCIISTRING, DataType.A_UTF8STRING]:
            code = code[::-1]

        # TODO: Apply bit mask.
        if bit_mask is not None:
            raise NotImplementedError(
                f"Don't know how to handle bit_mask={bit_mask}."
            )

        return code

    def _minimal_byte_length_of(self, internal_value: Union[bytes, str]) -> int:
        """Helper method to get the minimal byte length.
        (needed for LeadingLength- and MinMaxLengthType)
        """
        # A_BYTEFIELD, A_ASCIISTRING, A_UNICODE2STRING, A_UTF8STRING
        if self.base_data_type == DataType.A_BYTEFIELD:
            byte_length = len(internal_value)
        elif self.base_data_type in [DataType.A_ASCIISTRING, DataType.A_UTF8STRING]:
            assert isinstance(internal_value, str)
            # TODO: Handle different encodings
            byte_length = len(bytes(internal_value, "utf-8"))
        elif self.base_data_type == DataType.A_UNICODE2STRING:
            assert isinstance(internal_value, str)
            byte_length = len(bytes(internal_value, "utf-16-le"))
            assert byte_length % 2 == 0, (f"The bit length of A_UNICODE2STRING must"
                                          f" be a multiple of 16 but is {8*byte_length}")
        return byte_length

    @abc.abstractclassmethod
    def convert_internal_to_bytes(self, internal_value, encode_state: EncodeState, bit_position: int) -> bytes:
        """Encode the internal value.

        Parameters
        ----------
        internal_value : python type corresponding to self.base_data_type
            the value to be encoded
        bit_position : int 

        length_keys : Dict[str, int]
            mapping from ID (of the length key) to bit length
            (only needed for ParamLengthInfoType) 
        """
        pass

    @abc.abstractclassmethod
    def convert_bytes_to_internal(self, decode_state: DecodeState, bit_position: int = 0):
        """Decode the parameter value from the coded message.

        Parameters
        ----------
        decode_state : DecodeState
            The decoding state

        Returns
        -------
        str or int or bytes or bytearray or dict
            the decoded parameter value
        int
            the next byte position after the extracted parameter
        """
        pass


class LeadingLengthInfoType(DiagCodedType):
    def __init__(self,
                 base_data_type,
                 bit_length,
                 base_type_encoding=None,
                 is_highlow_byte_order=True):
        super().__init__(base_data_type,
                         dct_type="LEADING-LENGTH-INFO-TYPE",
                         base_type_encoding=base_type_encoding,
                         is_highlow_byte_order=is_highlow_byte_order)
        self.bit_length = bit_length
        assert self.bit_length > 0, "A Leading length info type with bit length == 0 does not make sense."
        assert self.base_data_type in [DataType.A_BYTEFIELD, DataType.A_ASCIISTRING, DataType.A_UNICODE2STRING, DataType.A_UTF8STRING], (
            f"A leading length info type cannot have the base data type {self.base_data_type}."
        )

    def convert_internal_to_bytes(self, internal_value, encode_state: EncodeState, bit_position: int) -> Union[bytes, bytearray]:
        byte_length = self._minimal_byte_length_of(internal_value)

        length_byte = self._to_bytes(byte_length,
                                     bit_position=bit_position,
                                     bit_length=self.bit_length,
                                     base_data_type=DataType.A_UINT32,
                                     is_highlow_byte_order=self.is_highlow_byte_order)

        value_byte = self._to_bytes(internal_value,
                                    bit_position=0,
                                    bit_length=8*byte_length,
                                    base_data_type=self.base_data_type,
                                    is_highlow_byte_order=self.is_highlow_byte_order)

        return length_byte + value_byte

    def convert_bytes_to_internal(self, decode_state: DecodeState, bit_position: int = 0):
        coded_message = decode_state.coded_message

        # Extract length of the parameter value
        byte_length, byte_position = self._extract_internal(coded_message=coded_message,
                                                            byte_position=decode_state.next_byte_position,
                                                            bit_position=bit_position,
                                                            bit_length=self.bit_length,
                                                            base_data_type=DataType.A_UINT32,  # length is an integer
                                                            is_highlow_byte_order=self.is_highlow_byte_order)

        # Extract actual value
        # TODO: The returned value is None if the byte_length is 0. Maybe change it
        #       to some default value like an empty bytearray() or 0?
        value, next_byte_position = self._extract_internal(coded_message=coded_message,
                                                           byte_position=byte_position,
                                                           bit_position=0,
                                                           bit_length=8*byte_length,
                                                           base_data_type=self.base_data_type,
                                                           is_highlow_byte_order=self.is_highlow_byte_order)

        return value, next_byte_position

    def __repr__(self) -> str:
        repr_str = f"LeadingLengthInfoType(base_data_type='{self.base_data_type}', bit_length={self.bit_length}"
        if self.base_type_encoding is not None:
            repr_str += f", base_type_encoding={self.base_type_encoding}"
        if not self.is_highlow_byte_order:
            repr_str += f", is_highlow_byte_order={self.is_highlow_byte_order}"
        return repr_str + ")"

    def __str__(self) -> str:
        return self.__repr__()


class MinMaxLengthType(DiagCodedType):
    def __init__(self,
                 base_data_type,
                 min_length,
                 termination,
                 max_length=None,
                 base_type_encoding=None,
                 is_highlow_byte_order=True):
        super().__init__(base_data_type,
                         dct_type="MIN-MAX-LENGTH-TYPE",
                         base_type_encoding=base_type_encoding,
                         is_highlow_byte_order=is_highlow_byte_order)
        assert max_length is None or min_length <= max_length
        self.min_length = min_length
        self.max_length = max_length
        self.termination = termination

        assert self.base_data_type in [DataType.A_BYTEFIELD, DataType.A_ASCIISTRING, DataType.A_UNICODE2STRING, DataType.A_UTF8STRING], (
            f"A min-max length type cannot have the base data type {self.base_data_type}."
        )
        assert self.termination in ["ZERO", "HEX-FF", "END-OF-PDU"], (
            f"A min-max length type cannot have the termination {self.termination}"
        )

    def __termination_character(self):
        """Returns the termination character or None if it isn't defined."""
        # The termination character is actually not specified by ASAM
        # for A_BYTEFIELD but I assume it is only one byte.
        termination_char = None
        if self.termination == "ZERO":
            if self.base_data_type not in [DataType.A_UNICODE2STRING]:
                termination_char = bytes([0x0])
            else:
                termination_char = bytes([0x0, 0x0])
        elif self.termination == "HEX-FF":
            if self.base_data_type not in [DataType.A_UNICODE2STRING]:
                termination_char = bytes([0xff])
            else:
                termination_char = bytes([0xff, 0xff])
        return termination_char

    def convert_internal_to_bytes(self, internal_value, encode_state: EncodeState, bit_position: int) -> bytes:
        byte_length = self._minimal_byte_length_of(internal_value)

        # The coded value must have at least length min_length
        if byte_length < self.min_length:
            raise EncodeError(f"The internal value {internal_value} is only {byte_length} bytes"
                              f" long but the min length is {self.min_length}")
        # The coded value must not have a length greater than max_length
        if self.max_length and byte_length > self.max_length:
            raise EncodeError(f"The internal value {internal_value} requires {byte_length}"
                              f" bytes, but the max length is {self.max_length}")

        value_byte = self._to_bytes(internal_value,
                                    bit_position=0,
                                    bit_length=8*byte_length,
                                    base_data_type=self.base_data_type,
                                    is_highlow_byte_order=self.is_highlow_byte_order)

        if encode_state.is_end_of_pdu or byte_length == self.max_length:
            # All termination types may be ended by the PDU
            return value_byte
        else:
            termination_char = self.__termination_character()
            if self.termination == 'END-OF-PDU':
                termination_char = bytes()
            assert termination_char is not None, \
                (f"MinMaxLengthType with termination {self.termination}"
                 f"(min: {self.min_length}, max: {self.max_length}) failed encoding {internal_value}")
            return value_byte + termination_char

    def convert_bytes_to_internal(self, decode_state: DecodeState, bit_position: int = 0):
        if decode_state.next_byte_position + self.min_length > len(decode_state.coded_message):
            raise DecodeError("The PDU ended before min length was reached.")

        coded_message = decode_state.coded_message
        byte_position = decode_state.next_byte_position
        termination_char = self.__termination_character()

        # If no termination char is found, this is the next byte after the parameter.
        max_termination_byte = len(coded_message)
        if self.max_length is not None:
            max_termination_byte = min(max_termination_byte, byte_position + self.max_length)

        if self.termination != "END-OF-PDU":
            # The parameter either ends after max length, at the end of the PDU
            # or if a termination character is found.
            char_length = len(termination_char)  # either 1 or 2

            termination_byte = byte_position + self.min_length
            found_char = False
            # Search the termination character
            while termination_byte < max_termination_byte and not found_char:
                found_char = coded_message[termination_byte: termination_byte +
                                           char_length] == termination_char
                if not found_char:
                    termination_byte += char_length

            byte_length = termination_byte - byte_position

            # Extract the value
            value, byte = self._extract_internal(decode_state.coded_message,
                                                 byte_position=byte_position,
                                                 bit_position=bit_position,
                                                 bit_length=8*byte_length,
                                                 base_data_type=self.base_data_type,
                                                 is_highlow_byte_order=self.is_highlow_byte_order)
            assert byte == termination_byte

            # next byte starts after the termination character
            next_byte_position = byte+char_length if found_char else byte
            return value, next_byte_position
        else:
            # If termination == "END-OF-PDU", the parameter ends after max_length
            # or at the end of the PDU.
            byte_length = max_termination_byte - byte_position

            value, byte = self._extract_internal(decode_state.coded_message,
                                                 byte_position=byte_position,
                                                 bit_position=bit_position,
                                                 bit_length=8*byte_length,
                                                 base_data_type=self.base_data_type,
                                                 is_highlow_byte_order=self.is_highlow_byte_order)
            return value, byte

    def __repr__(self) -> str:
        repr_str = f"MinMaxLengthType(base_data_type='{self.base_data_type}', min_length={self.min_length}"
        if self.max_length is not None:
            repr_str += f", base_type_encoding={self.max_length}"
        if self.termination is not None:
            repr_str += f", termination={self.termination}"
        if self.base_type_encoding is not None:
            repr_str += f", base_type_encoding={self.base_type_encoding}"
        if not self.is_highlow_byte_order:
            repr_str += f", is_highlow_byte_order={self.is_highlow_byte_order}"
        return repr_str + ")"

    def __str__(self) -> str:
        return self.__repr__()


class ParamLengthInfoType(DiagCodedType):
    def __init__(self,
                 base_data_type: Union[str, DataType],
                 length_key_ref: str,
                 base_type_encoding=None,
                 is_highlow_byte_order=True):
        super().__init__(base_data_type,
                         dct_type="PARAM-LENGTH-INFO-TYPE",
                         base_type_encoding=base_type_encoding,
                         is_highlow_byte_order=is_highlow_byte_order)
        self.length_key_ref = length_key_ref

    def convert_internal_to_bytes(self, internal_value, encode_state: EncodeState, bit_position: int) -> bytes:
        bit_length = encode_state.length_keys[self.length_key_ref]

        return self._to_bytes(internal_value,
                              bit_position=bit_position,
                              bit_length=bit_length,
                              base_data_type=self.base_data_type,
                              is_highlow_byte_order=self.is_highlow_byte_order)

    def convert_bytes_to_internal(self, decode_state: DecodeState, bit_position: int = 0):
        # Find length key with matching ID.
        bit_length = None
        for parameter, value in decode_state.parameter_value_pairs:
            # if isinstance(param_value.parameter, LengthKeyParameter) would be prettier,
            # but leads to cyclic import...
            if parameter.parameter_type == "LENGTH-KEY" \
                    and parameter.id == self.length_key_ref:
                # The bit length of the parameter to be extracted is given by the length key.
                bit_length = value
                break

        assert value is not None, f"Did not find any length key with ID {self.length_key_ref}"

        # Extract the internal value and return.
        return self._extract_internal(decode_state.coded_message,
                                      decode_state.next_byte_position,
                                      bit_position,
                                      bit_length,
                                      self.base_data_type,
                                      self.is_highlow_byte_order)

    def __repr__(self) -> str:
        repr_str = f"ParamLengthInfoType(base_data_type='{self.base_data_type}', length_key_ref={self.length_key_ref}"
        if self.base_type_encoding is not None:
            repr_str += f", base_type_encoding={self.base_type_encoding}"
        if not self.is_highlow_byte_order:
            repr_str += f", is_highlow_byte_order={self.is_highlow_byte_order}"
        return repr_str + ")"

    def __str__(self) -> str:
        return self.__repr__()


class StandardLengthType(DiagCodedType):

    def __init__(self,
                 base_data_type: Union[str, DataType],
                 bit_length: int,
                 bit_mask=None,
                 condensed: bool = False,
                 base_type_encoding=None,
                 is_highlow_byte_order=True):
        super().__init__(base_data_type,
                         dct_type="STANDARD-LENGTH-TYPE",
                         base_type_encoding=base_type_encoding,
                         is_highlow_byte_order=is_highlow_byte_order)
        self.bit_length = bit_length
        self.bit_mask = bit_mask
        self.condensed = condensed

    def convert_internal_to_bytes(self, internal_value, encode_state: EncodeState, bit_position: int) -> bytes:
        return self._to_bytes(internal_value,
                              bit_position,
                              self.bit_length,
                              self.base_data_type,
                              is_highlow_byte_order=self.is_highlow_byte_order,
                              bit_mask=self.bit_mask)

    def convert_bytes_to_internal(self, decode_state: DecodeState, bit_position: int = 0):
        return self._extract_internal(decode_state.coded_message,
                                      decode_state.next_byte_position,
                                      bit_position,
                                      self.bit_length,
                                      self.base_data_type,
                                      self.is_highlow_byte_order,
                                      bit_mask=self.bit_mask)

    def __repr__(self) -> str:
        repr_str = f"StandardLengthType(base_data_type='{self.base_data_type}', bit_length={self.bit_length}"
        if self.bit_mask is not None:
            repr_str += f", bit_mask={self.bit_mask}"
        if self.condensed:
            repr_str += f", condensed={self.condensed}"
        if self.base_type_encoding is not None:
            repr_str += f", base_type_encoding={self.base_type_encoding}"
        if not self.is_highlow_byte_order:
            repr_str += f", is_highlow_byte_order={self.is_highlow_byte_order}"
        return repr_str + ")"

    def __str__(self) -> str:
        return self.__repr__()


def read_diag_coded_type_from_odx(et_element):
    base_type_encoding = et_element.get("BASE-TYPE-ENCODING")

    base_data_type = et_element.get("BASE-DATA-TYPE")
    assert base_data_type in ["A_INT32", "A_UINT32", "A_FLOAT32", "A_FLOAT64",
                              "A_ASCIISTRING", "A_UTF8STRING", "A_UNICODE2STRING", "A_BYTEFIELD"]

    if et_element.get("IS-HIGHLOW-BYTE-ORDER") is not None:
        assert et_element.get("IS-HIGHLOW-BYTE-ORDER") in ["true", "false"]
        is_highlow_byte_order = et_element.get(
            "IS-HIGHLOW-BYTE-ORDER") == "true"
        logger.debug(
            f"HIGH-LOW-BYTE-ORDER set to {is_highlow_byte_order}")
    else:
        is_highlow_byte_order = True

    dct_type = et_element.get(f"{xsi}type")
    bit_length = None
    if dct_type == "LEADING-LENGTH-INFO-TYPE":
        bit_length = int(et_element.find("BIT-LENGTH").text)
        return LeadingLengthInfoType(base_data_type,
                                     bit_length=bit_length,
                                     base_type_encoding=base_type_encoding,
                                     is_highlow_byte_order=is_highlow_byte_order)
    elif dct_type == "MIN-MAX-LENGTH-TYPE":
        min_length = int(et_element.find("MIN-LENGTH").text)
        max_length = None
        if et_element.find("MAX-LENGTH"):
            max_length = int(et_element.find("MAX-LENGTH").text)
        termination = et_element.get("TERMINATION")

        return MinMaxLengthType(base_data_type,
                                min_length=min_length,
                                max_length=max_length,
                                termination=termination,
                                base_type_encoding=base_type_encoding,
                                is_highlow_byte_order=is_highlow_byte_order)
    elif dct_type == "PARAM-LENGTH-INFO-TYPE":
        length_key_ref = et_element.find(
            "LENGTH-KEY-REF").get("ID-REF")

        return ParamLengthInfoType(base_data_type,
                                   length_key_ref,
                                   base_type_encoding=base_type_encoding,
                                   is_highlow_byte_order=is_highlow_byte_order)
    elif dct_type == "STANDARD-LENGTH-TYPE":
        bit_length = int(et_element.find("BIT-LENGTH").text)
        bit_mask = None
        if et_element.find("BIT-MASK"):
            bit_mask = et_element.find("BIT-MASK").text
        condensed = et_element.get("CONDENSED") == "true"
        return StandardLengthType(base_data_type,
                                  bit_length,
                                  bit_mask=bit_mask,
                                  condensed=condensed,
                                  base_type_encoding=base_type_encoding,
                                  is_highlow_byte_order=is_highlow_byte_order)
    raise NotImplementedError(f"I do not know the diag-coded-type {dct_type}")



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from typing import Optional, Union

from ..odxtypes import DataType

from .compumethodbase import CompuMethod
from .limit import IntervalType, Limit


class LinearCompuMethod(CompuMethod):
    """Represents the decoding function d(y) = (offset + factor * y) / denominator
    where d(y) is the physical value and y is the internal value.

    Examples
    --------

    Define the decoding function `d(y) = 4+2*y` (or equivalent encoding `e(x) = floor((x-4)/2)`)
    on all integers `y` in the range -10..10 (and `x` in -16..25).

    ```python
    method = LinearCompuMethod(
        offset=4,
        factor=2,
        internal_type=DataType.A_INT32,
        physical_type=DataType.A_INT32,
        internal_lower_limit = Limit(-10, IntervalType.CLOSED),
        internal_upper_limit = Limit(11, IntervalType.OPEN)
    )
    ```

    Decode an internal value:

    ```python
    >>> method.convert_internal_to_physical(6)  # == 4+2*6
    16
    ```

    Encode a physical value:

    ```python
    >>> method.convert_physical_to_internal(6)  # == 6/2-2
    1
    ```

    Get physical limits:

    ```python
    >>> method.physical_lower_limit
    Limit(value=-16, interval_type=IntervalType.CLOSED)
    >>> method.physical_upper_limit
    Limit(value=26, interval_type=IntervalType.OPEN)
    ```

    (Note that there may be additional restrictions to valid physical values by the surrounding data object prop.
    For example, limits given by the bit length are not considered in the compu method.)
    """

    def __init__(self,
                 offset,
                 factor,
                 internal_type: Union[DataType, str],
                 physical_type: Union[DataType, str],
                 denominator=1,
                 internal_lower_limit: Optional[Limit] = None,
                 internal_upper_limit: Optional[Limit] = None):
        super().__init__(internal_type, physical_type, "LINEAR")
        self.offset = offset
        self.factor = factor
        self.denominator = denominator

        self.internal_lower_limit = internal_lower_limit
        if internal_lower_limit is None or internal_lower_limit.interval_type == IntervalType.INFINITE:
            self.internal_lower_limit = Limit(float("-inf"),
                                              IntervalType.INFINITE)

        self.internal_upper_limit = internal_upper_limit
        if internal_upper_limit is None or internal_upper_limit.interval_type == IntervalType.INFINITE:
            self.internal_upper_limit = Limit(float("inf"),
                                              IntervalType.INFINITE)

        assert self.internal_lower_limit is not None and self.internal_upper_limit is not None
        assert denominator > 0 and isinstance(denominator, (int, float))

        self.__compute_physical_limits()

    @property
    def physical_lower_limit(self):
        return self._physical_lower_limit

    @property
    def physical_upper_limit(self):
        return self._physical_upper_limit

    def __compute_physical_limits(self):
        """Computes the physical limits and stores them in the properties
        self._physical_lower_limit and self._physical_upper_limit.
        This method is only called during the initialization of a LinearCompuMethod.
        """
        def convert_to_limit_to_physical(limit: Limit, is_upper_limit: bool):
            """Helper method

            Parameters:

            limit
                the internal limit to be converted to a physical limit
            is_upper_limit
                True iff limit is the internal upper limit
            """
            assert isinstance(limit.value, (int, float))
            if limit.interval_type == IntervalType.INFINITE:
                return limit
            elif limit.interval_type == limit.interval_type.OPEN and self.internal_type.as_python_type() == int:
                closed_limit = limit.value - 1 if is_upper_limit else limit.value + 1
                return Limit(
                    value=self._convert_internal_to_physical(closed_limit),
                    interval_type=IntervalType.CLOSED
                )
            else:
                return Limit(
                    value=self._convert_internal_to_physical(limit.value),
                    interval_type=limit.interval_type
                )

        if self.factor >= 0:
            self._physical_lower_limit = convert_to_limit_to_physical(
                self.internal_lower_limit, False)
            self._physical_upper_limit = convert_to_limit_to_physical(
                self.internal_upper_limit, True)
        else:
            # If the factor is negative, the lower and upper limit are swapped
            self._physical_lower_limit = convert_to_limit_to_physical(
                self.internal_upper_limit, True)
            self._physical_upper_limit = convert_to_limit_to_physical(
                self.internal_lower_limit, False)

        if self.physical_type == DataType.A_UINT32:
            # If the data type is unsigned, the physical lower limit should be at least 0.
            if self._physical_lower_limit.interval_type == IntervalType.INFINITE or self._physical_lower_limit.value < 0:
                self._physical_lower_limit = Limit(
                    value=0,
                    interval_type=IntervalType.CLOSED
                )

    def _convert_internal_to_physical(self, internal_value):
        if self.denominator is None:
            result = self.offset + self.factor * internal_value
        else:
            result = (self.offset + self.factor *
                      internal_value) / self.denominator

        if self.internal_type == DataType.A_FLOAT64 and self.physical_type in [DataType.A_INT32, DataType.A_UINT32]:
            result = round(result)
        return self.physical_type.make_from(result)

    def convert_internal_to_physical(self, internal_value) -> Union[int, float]:
        assert self.is_valid_internal_value(internal_value)
        return self._convert_internal_to_physical(internal_value)

    def convert_physical_to_internal(self, physical_value):
        assert self.is_valid_physical_value(
            physical_value), f"physical value {physical_value} of type {type(physical_value)} is not valid. Expected type {self.physical_type} with internal range {self.internal_lower_limit} to {self.internal_upper_limit}"
        if self.denominator is None:
            result = (physical_value - self.offset) / self.factor
        else:
            result = ((physical_value * self.denominator) -
                      self.offset) / self.factor

        if self.physical_type == DataType.A_FLOAT64 and self.internal_type in [DataType.A_INT32, DataType.A_UINT32]:
            result = round(result)
        return self.internal_type.make_from(result)

    def is_valid_physical_value(self, physical_value):
        # Do type checks
        expected_type = self.physical_type.as_python_type()
        if expected_type == float and type(physical_value) not in [int, float]:
            return False
        elif expected_type != float and type(physical_value) != expected_type:
            return False

        # Compare to the limits
        if not self.physical_lower_limit.complies_to_lower(physical_value):
            return False
        if not self.physical_upper_limit.complies_to_upper(physical_value):
            return False
        return True

    def is_valid_internal_value(self, internal_value):
        expected_type = self.internal_type.as_python_type()
        if expected_type == float and type(internal_value) not in [int, float]:
            return False
        elif expected_type != float and type(internal_value) != expected_type:
            return False

        if not self.internal_lower_limit.complies_to_lower(internal_value):
            return False
        if not self.internal_upper_limit.complies_to_upper(internal_value):
            return False

        return True



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from typing import NamedTuple, Optional, Union

from .compurationalcoeffs import CompuRationalCoeffs
from .limit import Limit


class CompuScale(NamedTuple):
    """A COMPU-SCALE represents one value range of a COMPU-METHOD.

    Example:

    For a TEXTTABLE compu method a compu scale within COMPU-INTERNAL-TO-PHYS
    can be defined with
    ```
    scale = CompuScale(
        short_label="example_label", # optional: provide a label
        description="<p>fancy description</p>", # optional: provide a description
        lower_limit=Limit(0), # required: lower limit
        upper_limit=Limit(3), # required: upper limit
        compu_inverse_value=2, # required iff lower_limit != upper_limit
        compu_const="true", # required: physical value to be shown to the user
    )
    ```

    Almost all attributes are optional but there are compu-method-specific restrictions.
    E.g., lower_limit must always be defined unless the COMPU-METHOD is of CATEGORY LINEAR or RAT-FUNC.
    Either `compu_const` or `compu_rational_coeffs` must be defined but never both.
    """
    short_label: Optional[str] = None
    description: Optional[str] = None
    lower_limit: Optional[Limit] = None
    upper_limit: Optional[Limit] = None
    compu_inverse_value: Optional[Union[float, str, bytearray]] = None
    compu_const: Optional[Union[float, str]] = None
    compu_rational_coeffs: Optional[CompuRationalCoeffs] = None
